# coding=utf-8
# Copyright (C) Alex Urban (2018-)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Utilties for writing omega scan HTML pages
"""

import os
import numpy
from functools import wraps
from collections import OrderedDict

from MarkupPy import markup

from gwpy.table import Table
from gwpy.time import tconvert

from ..io import html as htmlio

__author__ = 'Alex Urban <alexander.urban@ligo.org>'
__credit__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


# -- HTML construction --------------------------------------------------------

def update_toc(toc, channel, name='GW'):
    """Add a channel to the page table of contents

    Parameters
    ----------
    toc : `dict`
        dictionary used as table of contents for a bootstrap page

    channel : `OmegaChannel`
        channel to be added to `toc`

    name : `str`, optional
        name of a channel's block, default: `'GW'`

    Returns
    -------
    out : `dict`
        the updated dictionary
    """
    out = toc
    try:  # update analyzed dict
        out[channel.section]['channels'].append(channel)
    except KeyError:
        out[channel.section] = {
            'name': name,
            'channels': [channel]}
    return out


def navbar(ifo, gpstime, toc={}):
    """Initialise a new `markup.page`

    Parameters
    ----------
    ifo : `str`
        the interferometer prefix

    gpstime : `float`
        the central GPS time of the analysis

    toc : `dict`
        metadata dictionary for navbar table of contents

    Returns
    -------
    page : `markup.page`
        the structured markup to open an HTML document
    """
    (brand, class_) = htmlio.get_brand(ifo, '&Omega;-scan',
                                       gpstime, about='about')
    # channel navigation
    links = [str(gpstime), ['Summary', '#']]
    for key, block in toc.items():
        channels = [[c.name, '#%s' % c.name.lower().replace(':', '-')]
                    for c in block['channels']]
        links.append([key, [[block['name'], channels]]])
    # return navbar
    return htmlio.navbar(links, brand=brand, class_=class_)


def wrap_html(func):
    """Decorator to wrap a function with `init_page` and `close_page` calls
    This allows inner HTML methods to be written with minimal arguments
    and content, hopefully making things simpler
    """
    @wraps(func)
    def decorated_func(ifo, gpstime, *args, **kwargs):
        # set page init args
        initargs = {
            'title': '%s Qscan | %s' % (ifo, gpstime),
            'base': os.path.curdir,
        }
        for key in ['title', 'base']:
            if key in kwargs:
                initargs[key] = kwargs.pop(key)
        # find outdir
        outdir = kwargs.pop('outdir', initargs['base'])
        if not os.path.isdir(outdir):
            os.makedirs(outdir)
        # determine table of contents and refresh options
        toc = kwargs.pop('toc', {})
        refresh = kwargs.pop('refresh', False)
        # determine primary channel
        correlated = kwargs.pop('correlated', False)
        primary = kwargs.pop('primary', '%s:GDS-CALIB_STRAIN' % ifo)
        # write about page
        try:
            config = kwargs.pop('config')
        except KeyError:
            about = None
        else:
            iargs = initargs.copy()
            aboutdir = os.path.join(outdir, 'about')
            prog = kwargs.pop('prog', None)
            if iargs['base'] == os.path.curdir:
                iargs['base'] = os.path.pardir
            iargs['toc'] = toc
            about = write_about_page(
                ifo, gpstime, config, prog=prog, outdir=aboutdir, **iargs)
            if os.path.basename(about) == 'index.html':
                about = about[:-10]
        # open page
        banner = navbar(ifo, gpstime, toc=toc)
        page = htmlio.new_bootstrap_page(refresh=refresh, navbar=banner,
                                         **initargs)
        # write analysis summary
        # (but only on the main results page)
        if about:
            page.add(write_summary(ifo, gpstime, incomplete=refresh))
            write_summary_table(toc, correlated)
            if correlated:
                page.add(write_ranking(toc, primary))
            kwargs['context'] = ifo.lower()
        # write content
        page.div(id_='main')
        # insert inner html directly
        page.add(str(func(*args, **kwargs)))
        page.div.close()  # main
        # close page
        index = os.path.join(outdir, 'index.html')
        htmlio.close_page(page, index, about=about)
        return index
    return decorated_func


# -- Utilities ----------------------------------------------------------------

def toggle_link(plottype, channel, pranges):
    """Create a Bootstrap button object that toggles between plot types.

    Parameters
    ----------
    plottype : `str`
        the type of plot to toggle toward

    channel : `OmegaChannel`
        the channel object corresponding to the plots shown

    pranges : `list` of `int`
        a list of ranges for the time axis of each plot

    Returns
    -------
    page : `page`
        a markup page object
    """
    text = plottype.split('_')[1]
    pstrings = ["'%s'" % p for p in pranges]
    chanstring = channel.name.replace('-', '_').replace(':', '-')
    captions = ["'%s'" % p.caption for p in channel.plots[plottype]]
    return markup.oneliner.a(
        text, class_='dropdown-item image-switch', **{
            'data-captions': '[%s]' % ','.join(captions),
            'data-channel-name': chanstring,
            'data-image-dir': 'plots',
            'data-image-type': plottype,
            'data-t-ranges': '[%s]' % ','.join(pstrings),
        })


def write_summary_table(blocks, correlated, base=os.path.curdir):
    """Write a summary table in various formats for users to download

    Parameters
    ----------
    blocks : `dict` of `OmegaChannel`
        the channel blocks scanned in the analysis

    correlated : `bool`
        Boolean switch to determine if cross-correlation is included

    base : `str`
        the path for the `<base>` tag to link in the `<head>`
    """
    # record summary data for each channel
    channel, time, freq, Q, energy, snr = ([], [], [], [], [], [])
    if correlated:
        corr, stdev, delay = ([], [], [])
    for block in blocks.values():
        for chan in block['channels']:
            channel.append(chan.name)
            time.append(chan.t)
            freq.append(chan.f)
            Q.append(chan.Q)
            energy.append(chan.energy)
            snr.append(chan.snr)
            if correlated:
                corr.append(chan.corr)
                stdev.append(chan.stdev)
                delay.append(chan.delay)
    # store in a table
    if correlated:
        data = Table([channel, time, freq, Q, energy, snr, corr, stdev, delay],
                     names=('Channel', 'Central Time',
                            'Central Frequency (Hz)', 'Q', 'Energy', 'SNR',
                            'Correlation', 'Standard Deviation', 'Delay (ms)'))
    else:
        data = Table([channel, time, freq, Q, energy, snr], names=(
            'Channel', 'Central Time', 'Central Frequency (Hz)', 'Q', 'Energy',
            'SNR'))
    # write in several formats
    datadir = os.path.join(base, 'data')
    fname = os.path.join(datadir, 'summary')
    data.write(fname + '.txt', format='ascii', overwrite=True)
    data.write(fname + '.csv', format='csv', overwrite=True)
    data.write(fname + '.tex', format='latex', overwrite=True)


# -- Qscan HTML ---------------------------------------------------------------

def write_summary(
        ifo, gpstime, incomplete=False, header='Summary',
        tableclass='table table-sm table-hover'):
    """Write the Qscan analysis summary HTML

    Parameters
    ----------
    ifo : `str`
        the interferometer prefix

    gpstime : `float`
        the central GPS time of the analysis

    incomplete : `bool`
        boolean switch to determine whether the scan is still in progress

    header : `str`, optional
        the text for the section header (``<h2``>)

    tableclass : `str`, optional
        the ``class`` for the summary ``<table>``

    Returns
    -------
    page : `~MarkupPy.markup.page`
        the formatted markup object containing the analysis summary table
    """
    utc = tconvert(gpstime)
    page = markup.page()
    page.div(class_='banner')
    page.h2(header)
    page.div.close()  # banner
    page.div(class_='row')

    page.div(class_='col-md-5')
    page.table(class_=tableclass)
    # make table body
    page.tbody()
    page.tr()
    page.th('Interferometer', scope='row')
    page.td('%s (%s)' % (htmlio.OBSERVATORY_MAP[ifo]['name'], ifo))
    page.tr.close()
    page.tr()
    page.th('UTC Time', scope='row')
    page.td(str(utc))
    page.tr.close()
    page.tbody.close()
    # close table
    page.table.close()
    page.div.close()  # col-md-5

    # make summary table download button
    page.div(class_='col-sm-12 col-md-7')
    content = [(
        ext,
        'data/summary.{}'.format(ext),
        '{0}_{1}_summary.{2}'.format(ifo, gpstime, ext)
    ) for ext in ('txt', 'csv', 'tex')]
    page.add(htmlio.download_btn(
        content, btnclass='btn btn-{} dropdown-toggle'.format(ifo.lower()),
    ))
    page.div.close()  # col-md-7
    page.div.close()  # row

    # write alert
    if incomplete:
        page.add(htmlio.alert(
            ('<strong>Note</strong>: This scan is in progress, and will '
             'auto-refresh every 60 seconds until completion.'),
            context=ifo.lower()))
    return page()


def write_ranking(toc, primary, thresh=6.5,
                  tableclass='table table-sm table-hover table-bordered'):
    """Write a table of channels ranked by their similarity to the primary

    Parameters
    ----------
    toc : `dict`
        metadata dictionary for navbar table of contents

    primary : `str`
        the name of the primary channel

    thresh : `float`
        the minimum correlation amplitude for appearing in this table

    tableclass : `str`, optional
        the ``class`` for the summary ``<table>``

    Returns
    -------
    page : `~MarkupPy.markup.page`
        the formatted markup object containing the analysis summary table
    """
    # construct an ordered dict of channel entries
    entries = OrderedDict([
        ('Channel', numpy.array([c.name for b in toc.values()
                                 for c in b['channels']])),
        ('GPS Time', numpy.array([c.t for b in toc.values()
                                  for c in b['channels']])),
        ('Frequency', numpy.array([c.f for b in toc.values()
                                   for c in b['channels']])),
        ('Q', numpy.array([c.Q for b in toc.values()
                           for c in b['channels']])),
        ('Energy', numpy.array([c.energy for b in toc.values()
                                for c in b['channels']])),
        ('SNR', numpy.array([c.snr for b in toc.values()
                             for c in b['channels']])),
        ('Correlation', numpy.array([c.corr for b in toc.values()
                                     for c in b['channels']])),
        ('Delay', numpy.array([c.delay for b in toc.values()
                               for c in b['channels']]))
    ])
    std = numpy.array([c.stdev for b in toc.values()
                       for c in b['channels']])

    # identify the primary channel
    pind = numpy.nonzero(entries['Channel'] == primary)
    # sort by matched-filter correlation
    ind_sorted = numpy.argsort(entries['Correlation'])[::-1]
    if not numpy.array_equiv(ind_sorted[0], pind):
        # prepend the primary channel
        dind = numpy.nonzero(ind_sorted == pind)
        ind_sorted = numpy.delete(ind_sorted, dind)
        ind_sorted = numpy.insert(ind_sorted, 0, pind)

    # construct HTML table
    page = markup.page()
    page.div(class_='row')
    page.div(class_='col-md-12')
    aparams = {
        'title': 'Whitened timeseries of the primary channel, %s.' % primary,
        'class_': 'fancybox cis-link',
        'data-fancybox': 'gallery',
        'data-fancybox-group': 'images',
        'target': '_blank',
    }
    tlink = markup.oneliner.a(primary, href='plots/primary.png', **aparams)
    page.p('Below are the top 5 channels ranked by matched-filter correlation '
           'within 100 ms of %s.' % tlink)
    # range over channels
    k = 0
    data = []
    for i in ind_sorted:
        if k > 5:
            break
        # reject channels with too high a glitch rate
        if (std[i] > 2) and (entries['Channel'][i] != primary):
            continue
        params = {
            'title': entries['Channel'][i],
            'href': '#%s' % entries['Channel'][i].lower().replace(':', '-'),
            'class_': 'cis-link',
        }
        row = [
            markup.oneliner.a(entries['Channel'][i], **params),
            str(entries['GPS Time'][i]),
            '%.1f Hz' % entries['Frequency'][i],
            str(entries['Q'][i]),
            str(entries['Energy'][i]),
            str(entries['SNR'][i]),
        ]
        if entries['Channel'][i] == primary:
            row.extend(['&mdash;', '&mdash;'])
        else:
            row.extend([str(entries['Correlation'][i]),
                        '%s ms' % entries['Delay'][i]])
        # increment counter
        data.append(row)
        k += 1
    page.add(htmlio.table(
        headers=entries.keys(), data=data, separator='\n', table=tableclass))
    page.div.close()  # col-md-12
    page.div.close()  # row
    return page()


def write_block(blockkey, block, context,
                tableclass='table table-sm table-hover '
                           'table-bordered  d-none d-lg-table'):
    """Write the HTML summary for a specific block of channels

    Parameters
    ----------
    blockkey: `str`
        the key labeling the channel block

    block : `dict` of `OmegaChannel`
        a list of channels and their analysis attributes

    context : `str`
        the type of Bootstrap ``<card>`` object to use, color-coded by GWO
        standards (must be one of 'h1', 'l1', 'v1', 'k1', 'i1', 'g1', or
        'network')

    tableclass : `str`, optional
        the ``class`` for the summary ``<table>``

    Returns
    -------
    page : `~MarkupPy.markup.page`
        the formatted HTML for this block
    """
    page = markup.page()
    page.div(class_='card card-%s mb-5 shadow-sm' % context)
    # -- make heading
    page.div(class_='card-header pb-0')
    page.h5(': '.join([blockkey, block['name']]), class_='card-title')
    page.div.close()  # card-header pb-0

    # -- make body
    page.div(class_='card-body')
    page.div(class_='list-group')

    # -- range over channels in this block
    for i, channel in enumerate(block['channels']):
        page.div(class_='list-group-item flex-column align-items-start')

        # channel name
        chanid = channel.name.lower().replace(':', '-')
        page.h5(htmlio.cis_link(channel.name), class_='card-title', id_=chanid)

        # summary table
        page.div(class_='row')
        page.div(class_='col-sm-12 col-md-7')
        try:
            columns = ['GPS Time', 'Frequency', 'Q', 'Energy', 'SNR',
                       'Correlation', 'Delay']
            entries = [[str(channel.t), '%s Hz' % channel.f, str(channel.Q),
                        str(channel.energy), str(channel.snr),
                        str(channel.corr), '%s ms' % channel.delay]]
        except AttributeError:
            columns = ['GPS Time', 'Frequency', 'Q', 'Energy', 'SNR']
            entries = [[str(channel.t), '%s Hz' % channel.f, str(channel.Q),
                        str(channel.energy), str(channel.snr)]]
        page.add(
            htmlio.table(columns, entries, separator='\n', table=tableclass))
        page.div.close()  # col-sm-12 col-md-7

        # plot toggle buttons
        page.div(class_='col-sm-12 col-md-5')
        page.div(class_='btn-group flex-wrap', role='group')
        for ptitle, pclass, ptypes in [
            ('Timeseries', 'timeseries', ('raw', 'highpassed', 'whitened')),
            ('Spectrogram', 'qscan', ('highpassed', 'whitened', 'autoscaled')),
            ('Eventgram', 'eventgram', (
                'highpassed', 'whitened', 'autoscaled')),
        ]:
            _id = 'btnGroup{0}{1}'.format(pclass.title(), i)
            page.div(class_='btn-group', role='group')
            page.button(ptitle, id_=_id, type='button',
                        class_='btn btn-%s dropdown-toggle' % context,
                        **{'data-toggle': 'dropdown'})
            page.div(class_='dropdown-menu shadow', **{'aria-labelledby': _id})
            for ptype in ptypes:
                page.add(toggle_link('{0}_{1}'.format(pclass, ptype),
                                     channel, channel.pranges))
            page.div.close()  # dropdown-menu
            page.div.close()  # btn-group
        page.div.close()  # btn-group
        page.div.close()  # col-sm-12 col-md-5
        page.div.close()  # row

        # plots
        page.add(htmlio.scaffold_plots(
            channel.plots['qscan_whitened'],
            nperrow=min(len(channel.pranges), 3)))

        page.div.close()  # list-group-item

    # close and return
    page.div.close()  # list-group
    page.div.close()  # card-body
    page.div.close()  # card
    return page()


# reminder: wrap_html automatically prepends the (ifo, gpstime) args,
# and at least the outdir kwarg, so you should include those in the docstring,
# but not in the actual function declaration - the decorator will take care of
# that for you.

@wrap_html
def write_qscan_page(blocks, context):
    """Write the Qscan results to HTML

    Parameters
    ----------
    ifo : `str`
        the prefix of the interferometer used in this analysis

    gpstime  : `float`
        the central GPS time of the analysis

    blocks : `dict` of `OmegaChannel`
        the channel blocks scanned in the analysis

    context : `str`, optional
        the type of Bootstrap ``<card>`` object to use, color-coded by
        GWO standard

    Returns
    -------
    index : `str`
        the path of the HTML written for this analysis
    """
    page = markup.page()
    page.div(class_='banner')
    page.h2('Channel details')
    page.div.close()  # banner
    for key, block in blocks.items():
        page.add(write_block(key, block, context))
    return page


@wrap_html
def write_null_page(reason, context='info'):
    """Write the Qscan results to HTML

    Parameters
    ----------
    ifo : `str`
        the prefix of the interferometer used in this analysis

    gpstime  : `float`
        the central GPS time of the analysis

    reason : `str`
        the explanation for this null result

    context : `str`
        type of bootstrap alert box to use, default: ``info``

    Returns
    -------
    index : `str`
        the path of the HTML written for this analysis
    """
    page = markup.page()
    page.add(htmlio.alert(reason, context=context, dismiss=False))
    return page


@wrap_html
def write_about_page(configfiles, prog=None):
    """Write a page explaining how a Qscan analysis was completed

    Parameters
    ----------
    ifo : `str`
        the prefix of the interferometer used in this analysis

    gpstime  : `float`
        the central GPS time of the analysis

    configfiles : `list` of `str`
        list of paths of the configuration files to embed

    prog : `str`, optional
        name of the program which produced this page, defaults to
        the script run on the command-line

    outdir : `str`, optional
        the output directory for the HTML

    Returns
    -------
    index : `str`
        the path of the HTML written for this analysis
    """
    # set up page
    return htmlio.about_this_page(
        configfiles[0] if len(configfiles) == 1 else configfiles,
        prog=prog,
    )



"""Tests for gwdetchar.lasso.__main__
"""
import pytest

import numpy as np

from gwpy.timeseries import TimeSeries

from .. import __main__ as lasso


__author__ = "Michael Lowry <michaeljohn.lowry@ligo.org>"


# global test objects

TEST_CHANNEL = "test_channel"
TEST_START = 0
TEST_END = TEST_START+1000


# # -- pytest fixtures -------------------------------------------------------
@pytest.fixture
def expected_ts():
    # construct random data to read
    data = np.linspace(TEST_START, TEST_END, num=TEST_END-TEST_START)
    return TimeSeries(data, t0=TEST_START, dt=1, channel=TEST_CHANNEL,
                      name=TEST_CHANNEL)


@pytest.fixture
def expected_ts_file(expected_ts, tmp_path):
    # write data to file and return that file
    outfile = tmp_path / "data.gwf"
    expected_ts.write(outfile, format='gwf')
    return outfile


# # -- unit tests -------------------------------------------------------------
def test_read(expected_ts, expected_ts_file):
    # test reading primary TimeSeries file
    actual_ts = lasso.get_primary_ts(
        filepath=expected_ts_file,
        channel=TEST_CHANNEL,
        start=TEST_START,
        end=TEST_END,
        cache=None,
        nproc=1)
    assert actual_ts.t0 == expected_ts.t0
    assert actual_ts.times[-1] == expected_ts.times[-1]
    assert actual_ts.sample_rate == expected_ts.sample_rate
    assert actual_ts.dx == expected_ts.dx
    np.testing.assert_array_equal(
        actual_ts.value,
        expected_ts.value,
        err_msg='read in data array does not match')



# coding=utf-8
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Simple command-line interface to gwdetchar.scattering

Given a specific GPS time of interest, this module scans through records of
optic motion and projects fringe frequencies due to optical scattering. For
those channels with fringes above a user-specified threshold, a plot is
created comparing the fringes to a high-resolution Q-scan spectrogram.

To identify broader time segments when scattering is likely in the first
place, please use the main command-line module:

python -m gwdetchar.scattering --help
"""

import os
import sys

from gwpy.time import to_gps

from .. import (cli, const)
from ..omega import highpass
from ..io.datafind import get_data

from . import (
    OPTIC_MOTION_CHANNELS,
    get_fringe_frequency,
)

from matplotlib import use
use('Agg')

# backend-dependent import
from . import plot  # noqa: E402

__author__ = 'Alex Urban <alexander.urban@ligo.org>'
__credits__ = 'Joshua Smith <joshua.smith@ligo.org>' \
              'Andrew Lundgren <andrew.lundgren>@ligo.org>' \
              'Duncan Macleod <duncan.macleod@ligo.org>'

# global variables

ASD_KW = {
    'method': 'median',
    'fftlength': 8,
    'overlap': 4,
}

MOTION_CHANNELS = [channel for key in OPTIC_MOTION_CHANNELS.keys()
                   for channel in OPTIC_MOTION_CHANNELS[key]]

PROG = ('python -m gwdetchar.scattering.simple' if sys.argv[0].endswith('.py')
        else os.path.basename(sys.argv[0]))
LOGGER = cli.logger(name=PROG.split('python -m ').pop())


# -- utilities ----------------------------------------------------------------

def _discover_data(primary, channels, start, end,
                   primary_frametype, aux_frametype):
    """Load timeseries data from local gravitational-wave frame files
    """
    hoft = get_data(
        primary,
        start=start - ASD_KW['overlap'],
        end=end + ASD_KW['overlap'],
        frametype=primary_frametype,
        verbose='Reading primary channel:'.rjust(30),
    )
    aux = get_data(
        channels,
        start=start,
        end=end,
        frametype=aux_frametype,
        verbose='Reading auxiliary sensors:'.rjust(30),
    )
    return (hoft, aux)


# -- parse command-line -------------------------------------------------------

def create_parser():
    """Create a command-line parser for this entry point
    """
    # initialize argument parser
    parser = cli.create_parser(
        prog=PROG,
        description=__doc__,
    )

    # required arguments
    parser.add_argument(
        'gpstime',
        type=to_gps,
        help='GPS time or datestring to analyze',
    )
    cli.add_ifo_option(
        parser,
        ifo=const.IFO,
    )

    # optional arguments
    parser.add_argument(
        '-d',
        '--duration',
        type=float,
        default=60,
        help='Duration (seconds) of analysis, default: 60',
    )
    parser.add_argument(
        '-t',
        '--frequency-threshold',
        type=float,
        default=15,
        help='critical fringe frequency threshold (Hz), '
             'default: %(default)s',
    )
    parser.add_argument(
        '-m',
        '--multipliers',
        default='1,2,4,8',
        help='harmonic numbers to plot projected motion for, '
             'should be given as a comma-separated list of '
             'numbers, default: %(default)s',
    )
    parser.add_argument(
        '-x',
        '--multiplier-for-threshold',
        type=int,
        default=4,
        help='frequency multiplier to use when applying '
             '--frequency-threshold, default: %(default)s',
    )
    parser.add_argument(
        '-w',
        '--primary-channel',
        default='GDS-CALIB_STRAIN',
        help='name of primary channel (without IFO prefix), '
             'default: %(default)s',
    )
    parser.add_argument(
        '-W',
        '--primary-frametype',
        default='{IFO}_HOFT_C00',
        help='frametype from which to read primary channel, '
             'default: %(default)s',
    )
    parser.add_argument(
        '-a',
        '--aux-frametype',
        default='{IFO}_R',
        help='frametype from which to read auxiliary channels, '
             'default: %(default)s',
    )
    parser.add_argument(
        '-o',
        '--output-dir',
        type=os.path.abspath,
        default=os.curdir,
        help='Output directory for analysis, default: %(default)s',
    )
    parser.add_argument(
        '-c',
        '--colormap',
        default='viridis',
        help='name of colormap to use, default: %(default)s',
    )

    # return the argument parser
    return parser


# -- main code block ----------------------------------------------------------

def main(args=None):
    """Run the simple version of the scattering command-line tool
    """
    parser = create_parser()
    args = parser.parse_args(args=args)

    # store arguments
    ifo = args.ifo
    gps = float(args.gpstime)
    gpsstart = gps - 0.5 * args.duration
    gpsend = gps + 0.5 * args.duration
    primary = ':'.join([ifo, args.primary_channel])
    channels = [':'.join([ifo, c]) for c in MOTION_CHANNELS]
    thresh = args.frequency_threshold
    multipliers = [int(x) for x in args.multipliers.split(',')]
    harmonic = args.multiplier_for_threshold
    if '{IFO}' in args.primary_frametype:
        args.primary_frametype = args.primary_frametype.format(IFO=ifo)
    if '{IFO}' in args.aux_frametype:
        args.aux_frametype = args.aux_frametype.format(IFO=ifo)
    LOGGER.info('{0} Scattering: {1}'.format(ifo, gps))

    # retrieve data
    (hoft, data) = _discover_data(primary, channels, gpsstart, gpsend,
                                  args.primary_frametype, args.aux_frametype)

    # set up spectrogram
    LOGGER.debug('Setting up a Q-scan spectrogram of {}'.format(primary))
    hoft = highpass(hoft, f_low=thresh).resample(256)
    qspecgram = hoft.q_transform(qrange=(4, 150), frange=(0, 60), gps=gps,
                                 fres=0.1, outseg=(gpsstart, gpsend), **ASD_KW)
    qspecgram.name = primary

    # process channels
    count = 0  # running count of plots written
    for channel in channels:
        LOGGER.info(' -- Processing {} -- '.format(channel))
        try:
            motion = data[channel].detrend().resample(128)
        except KeyError:
            LOGGER.warning('Skipping {}'.format(channel))
            continue
        # project scattering frequencies
        fringe = get_fringe_frequency(motion, multiplier=1)
        ind = fringe.argmax()
        fmax = fringe.value[ind]
        tmax = fringe.times.value[ind]
        LOGGER.debug('Maximum scatter frequency {0:.2f} Hz at GPS second '
                     '{1:.2f}'.format(fmax, tmax))
        if harmonic * fmax < thresh:
            LOGGER.warning('No significant evidence of scattering '
                           'found in {}'.format(channel))
            continue
        # plot spectrogram and fringe frequency
        output = os.path.join(
            args.output_dir,
            '%s-%s-%s-{}.png' % (
                channel.replace('-', '_').replace(':', '-', 1),
                gps, args.duration)
        )
        LOGGER.debug('Plotting spectra and projected fringe frequencies')
        plot.spectral_comparison(
            gps, qspecgram, fringe, output.format('comparison'), thresh=thresh,
            multipliers=multipliers, colormap=args.colormap)
        plot.spectral_overlay(
            gps, qspecgram, fringe, output.format('overlay'),
            multipliers=multipliers)
        LOGGER.info(' -- Channel complete -- ')
        count += 1  # increment counter
    LOGGER.info('{0:g} chanels plotted in {1}'.format(count, args.output_dir))


# -- run from command-line ----------------------------------------------------

if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright (C) Duncan Macleod (2015)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Utilties for HTML output
"""

import datetime
import json
import os
import subprocess
import sys

from collections import OrderedDict
from getpass import getuser
from io import StringIO
from operator import itemgetter
from pathlib import Path
from pytz import reference
from shutil import copyfile
from urllib.parse import urlparse

from MarkupPy import markup

from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import get_lexer_by_name

from gwpy.time import from_gps

from ..plot import plot_segments
from .._version import get_versions

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credit__ = 'Alex Urban <alexander.urban@ligo.org>'

# -- give context for ifo names

OBSERVATORY_MAP = {
    'G1': {
        'name': 'GEO',
        'links': OrderedDict([
            ('Network Summary Pages', 'https://ldas-jobs.ligo.caltech.edu/'
                                      '~detchar/summary/day')])
    },
    'H1': {
        'name': 'LIGO Hanford',
        'links': OrderedDict([
            ('LHO Summary Pages', 'https://ldas-jobs.ligo-wa.caltech.edu/'
                                  '~detchar/summary/day'),
            ('LHO Logbook', 'https://alog.ligo-wa.caltech.edu/aLOG')])
    },
    'I1': {
        'name': 'LIGO India',
        'links': OrderedDict([
            ('Network Summary Pages', 'https://ldas-jobs.ligo.caltech.edu/'
                                      '~detchar/summary/day')])
    },
    'K1': {
        'name': 'KAGRA',
        'links': OrderedDict([
            ('Network Summary Pages', 'https://ldas-jobs.ligo.caltech.edu/'
                                      '~detchar/summary/day'),
            ('KAGRA Logbook', 'http://klog.icrr.u-tokyo.ac.jp/osl')])
    },
    'L1': {
        'name': 'LIGO Livingston',
        'links': OrderedDict([
            ('LLO Summary Pages', 'https://ldas-jobs.ligo-la.caltech.edu/'
                                  '~detchar/summary/day'),
            ('LLO Logbook', 'https://alog.ligo-la.caltech.edu/aLOG')])
    },
    'V1': {
        'name': 'Virgo',
        'links': OrderedDict([
            ('Network Summary Pages', 'https://ldas-jobs.ligo.caltech.edu/'
                                      '~detchar/summary/day/'),
            ('Virgo Logbook', 'https://logbook.virgo-gw.eu/virgo')])
    },
    'Network': {
        'name': 'Multi-IFO',
        'links': OrderedDict([
            ('Network Summary Pages', 'https://ldas-jobs.ligo.caltech.edu/'
                                      '~detchar/summary/day'),
            ('LHO Logbook', 'https://alog.ligo-wa.caltech.edu/aLOG'),
            ('LLO Logbook', 'https://alog.ligo-la.caltech.edu/aLOG'),
            ('Virgo Logbook', 'https://logbook.virgo-gw.eu/virgo'),
            ('KAGRA Logbook', 'http://klog.icrr.u-tokyo.ac.jp/osl')])
    }
}

# -- HTML URLs

FONT_AWESOME_CSS = ('https://cdnjs.cloudflare.com/ajax/libs/'
                    'font-awesome/5.15.1/css/fontawesome.min.css')
FONT_AWESOME_SOLID_CSS = ('https://cdnjs.cloudflare.com/ajax/libs/'
                          'font-awesome/5.15.1/css/solid.min.css')

JQUERY_JS = 'https://code.jquery.com/jquery-3.5.1.min.js'
JQUERY_LAZY_JS = ('https://cdnjs.cloudflare.com/ajax/libs/jquery.lazy/'
                  '1.7.11/jquery.lazy.min.js')
BOOTSTRAP_JS = ('https://cdn.jsdelivr.net/npm/bootstrap@4.5.3/'
                'dist/js/bootstrap.bundle.min.js')
FANCYBOX_JS = ('https://cdnjs.cloudflare.com/ajax/libs/'
               'fancybox/3.5.7/jquery.fancybox.min.js')

GWBOOTSTRAP_CSS = ('https://cdn.jsdelivr.net/npm/gwbootstrap@1.3.2/'
                   'lib/gwbootstrap.min.css')
GWBOOTSTRAP_JS = ('https://cdn.jsdelivr.net/npm/gwbootstrap@1.3.2/'
                  'lib/gwbootstrap.min.js')

CSS_FILES = [
    FONT_AWESOME_CSS,
    FONT_AWESOME_SOLID_CSS,
    GWBOOTSTRAP_CSS,
]

JS_FILES = [
    JQUERY_JS,
    JQUERY_LAZY_JS,
    BOOTSTRAP_JS,
    FANCYBOX_JS,
    GWBOOTSTRAP_JS,
]

FORMATTER = HtmlFormatter(noclasses=True)


# -- Plot construction --------------------------------------------------------

class FancyPlot(object):
    """A helpful class of objects that coalesce image links and caption text
    for fancybox figures.

    Parameters
    ----------
    img : `str` or `FancyPlot`
        either a filename (including relative or absolute path) or another
        FancyPlot instance

    caption : `str`
        the text to be displayed in a fancybox as this figure's caption
    """
    def __init__(self, img, caption=None):
        if isinstance(img, FancyPlot):
            caption = caption if caption else img.caption
        self.img = str(img)
        self.caption = caption if caption else os.path.basename(self.img)

    def __str__(self):
        return self.img


# -- HTML construction --------------------------------------------------------

def _get_card_header(context):
    """Return the correct bootstrap-4 card-header class for the given context
    """
    if context == 'light':
        return 'card-header bg-light'
    else:
        return 'card-header text-white bg-%s' % context


def finalize_static_urls(static, base, cssfiles, jsfiles):
    """Finalise the necessary CSS and javascript files as URLS.

    The method parses the lists of files given, copying any local files into
    ``static`` as necessary to create resolvable URLs to include in the HTML
    ``<head>``.

    Parameters
    ----------
    static : `str`
        the target directory for the static files, will be created if
        necessary

    base : `str`
        the base directory of the website

    cssfiles : `list` of `str`
        the list of CSS files to include

    jsfiles : `list` of `str`
        the (complete) list of javascript files to include

    Returns
    -------
    cssurls : `list` of `str`
        the finalised list of CSS files

    jsurls : `list` of `str`
        the finalised list of javascript files
    """
    static = Path(static).resolve()
    base = Path(base).resolve()

    def _local_url(path):
        """Copy a filepath into the static dir if required
        """
        path = Path(path).resolve()
        # if file is already below static in the hierarchy, don't do anything
        if static in path.parents:
            return path.relative_to(base)
        # otherwise copy the file into static
        static.mkdir(parents=True, exist_ok=True)
        local = static / path.name
        copyfile(str(path), str(local))  # only need str for py<3.6
        return str(local.relative_to(base))

    # copy lists so that we can modify
    cssfiles = list(cssfiles)
    jsfiles = list(jsfiles)

    for flist in (cssfiles, jsfiles):
        for i, fn in enumerate(flist):
            url = urlparse(fn)
            if url.netloc in {"", "file"}:
                flist[i] = _local_url(fn)

    return cssfiles, jsfiles


def new_bootstrap_page(base=os.path.curdir, path=os.path.curdir, lang='en',
                       refresh=False, topbtn=True, navbar=None, **kwargs):
    """Create a new `~markup.page` with custom twitter bootstrap CSS and
    JS headers

    Parameters
    ----------
    base : `str`, optional
        relative path to the base directory, default: `.`

    path : `str`, optional
        path to directory where the page is located, default: `.`

    lang : `str`, optional
        language of the page, default: en

    refresh : `bool`, optional
        boolean switch to enable periodic page refresh, default: False

    topbtn : `bool`, optional
        boolean switch to include or exclude a floating button that scrolls
        to the top of the page, default: True

    navbar : `str`, optional
        HTML enconding of a floating navbar, will be ignored if not given,
        default: None

    Returns
    -------
    page : `~MarkupPy.markup.page`
        a populated HTML page object, which can be appended to as desired
    """
    # get kwargs with sensible defaults
    css = kwargs.pop('css', CSS_FILES)
    script = kwargs.pop('script', JS_FILES)
    # write CSS to static dir
    css, script = finalize_static_urls(
        os.path.join(path, 'static'),
        os.path.curdir,
        css,
        script,
    )
    # create page
    page = markup.page()
    page.header.append('<!DOCTYPE HTML>')
    page.html(lang=lang)
    page.head()
    if refresh:  # force-refresh if requested
        page.meta(http_equiv='refresh', content='60')
    # ensure nice formatting on most devices
    page.meta(http_equiv='Content-Type', content='text/html; charset=utf-8')
    page.meta(content='width=device-width, initial-scale=1.0', name='viewport')
    page.base(href=base)
    page._full = True
    # link files
    for f in css:
        page.link(href=f, rel='stylesheet', media='all')
    for f in script:
        page.script('', src=f, type='text/javascript')
    # add other attributes
    for key in kwargs:
        getattr(page, key)(kwargs[key])
    # finalize header
    page.head.close()
    # open body and container
    page.body()
    if topbtn:
        glyph = markup.oneliner.i('', class_='fas fa-arrow-up')
        page.button(glyph, title='Return to top',
                    class_='btn-float shadow', id_='top-btn')
    if navbar is not None:
        page.add(navbar)
    page.div(class_='container')
    return page


def navbar(links, class_='navbar navbar-expand-md fixed-top shadow-sm',
           brand=None, collapse=True):
    """Construct a navigation bar in bootstrap format

    Parameters
    ----------
    links : `list`
        list of either (text, URL) tuples or  :class:`~MarkupPy.markup.page`
        objects. Tuples will be written in ``<a>`` tags while `pages` will
        be copied directly. Both will be enclosed in a <li> element inside
        the navbar

    class_ : `str`, optional
        navbar object class, default: `'navbar navbar-expand-md fixed-top'`

    brand : `str`, `~MarkupPy.markup.page`, or `list`, optional
        branding for the navigation bar, default: None

    collapse : `bool`, optional
        whether to toggle all dropdown menus, default: True

    Returns
    -------
    page : :class:`~MarkupPy.markup.page`
        navbar HTML `page` object
    """
    if isinstance(brand, (tuple, list)):
        brand, help_ = brand
    else:
        help_ = None

    # page setup
    page = markup.page()
    page.twotags.append('nav')
    page.nav(class_=class_)
    page.div(class_='container-fluid')

    # add branding (generic non-collapsed content)
    if brand is not None:
        page.add(str(brand))

    # begin navbar proper
    if collapse:
        page.button(
            class_='navbar-toggler navbar-toggler-right', type_='button',
            **{'data-toggle': 'collapse', 'data-target': '.navbar-collapse'})
        page.span('', class_='navbar-toggler-icon')
        page.button.close()
        page.div(class_='collapse navbar-collapse justify-content-between')
    else:
        page.div()

    # ---- collapsable part (<div>) ----

    if links:
        page.ul(class_='nav navbar-nav mr-auto')
        for i, link in enumerate(links):
            if (isinstance(link, (list, tuple)) and
                    isinstance(link[1], str)):
                text, link = link
                page.li(class_='nav-item')
                page.a(text, href=link, class_='nav-link')
            elif (isinstance(link, (list, tuple)) and
                  isinstance(link[1], (list, tuple))):
                page.li(class_='nav-item dropdown')
                page.add(dropdown(*link))
            elif isinstance(link, str) and not link.startswith('<'):
                page.li(class_='nav-item navbar-text')
                page.add(link)
            else:
                page.li(class_='nav-item')
                page.add(str(link))
            page.li.close()
        page.ul.close()  # nav navbar-nav mr-auto

        if help_ is not None:
            page.add(str(help_))

    page.div.close()  # collapse navbar-collapse
    page.div.close()  # container-fluid
    page.nav.close()
    return page()


def dropdown(text, links, active=None, class_='nav-link dropdown-toggle'):
    """Construct a dropdown menu in bootstrap format

    Parameters
    ----------
    text : `str`
        dropdown menu header

    links : `list`
        list of (Link text, linkurl) tuples or dict of such tuples for
        grouped dropdowns

    active : `int` or `list` of `ints`, optional
        collection of links to make active, default: None

    class_ : `str`, optional
        object class of the dropdown menu, default: `'dropdown-toggle'`

    Returns
    -------
    page : :class:`~MarkupPy.markup.page`
        HTML element with the following grammar:

        .. code:: html

           <a>text</a>
           <ul>
               <li>link</li>
               <li>link</li>
           </ul>
    """
    def has_columns(items_):
        return (isinstance(items_, (tuple, list)) and len(items_)
                and isinstance(items_[1], (tuple, list)))

    def has_open_row(page_):
        tags = page_.content
        divs = [t for t in tags if ('<div' in t)]
        cldivs = [t for t in tags if ('</div' in t)]
        return len(cldivs) != len(divs) - 1

    page = markup.page()
    page.a(text, href='#', class_=class_, role='button',
           **{'data-toggle': 'dropdown'})

    # work out columns
    ngroup = sum([has_columns(x) for x in links])
    ncol = min(ngroup, 4) or 1
    page.div(class_='dropdown-menu dropdown-%d-col shadow' % ncol)

    # dropdown elements
    for i, link in enumerate(links):
        # handle active links
        if isinstance(active, int) and i == active:
            active_ = True
        elif isinstance(active, (list, tuple)) and i == active[0]:
            active_ = active[1]
        else:
            active_ = False
        # handle multi-column
        if has_columns(link):
            if not has_open_row(page):
                page.div(class_='row')
            dropdown_link(page, link, active=active_,
                          class_='col-sm-12 col-md-%d' % (12 // ncol))
        else:
            dropdown_link(page, link, active=active_)

    # close and return
    if has_open_row(page):
        page.div.close()  # row
    page.div.close()  # dropdown-menu
    return page()


def dropdown_link(page, link, active=False, class_=''):
    """Format links within a dropdown menu

    Parameters
    ----------
    page : `~MarkupPy.markup.page`
        a `page` object to format in-place

    link : `~MarkupPy.markup.page` or `list`
        the link(s) to format

    active : `bool`, optional
        boolean switch to enable (`True`) or disable (`False`) an active link,
        default: `False`

    class_ : `str`, optional
        object class of the link, default: `''`
    """
    if link in [None, '']:
        page.div('', class_='dropdown-divider')
    elif isinstance(link, (tuple, list)):
        if isinstance(link[1], (tuple, list)):
            page.div(class_=class_)
            page.h6(link[0], class_='dropdown-header')
            for j, link2 in enumerate(link[1]):
                dropdown_link(page, link2,
                              active=(type(active) is int and active == j))
            page.div.close()
        else:
            page.a(link[0], href=link[1],
                   class_=('dropdown-item active' if active is True
                           else 'dropdown-item'))
    elif link is not None:
        page.add(str(link))


def get_brand(ifo, name, gps, about=None):
    """Return a brand for navigation bar formatting

    Parameters
    ----------
    ifo : `str`
        interferometer prefix for color-coding, e.g. `'L1'`

    name : `str`
        name of the analysis, e.g. `'Scattering'`

    gps : `float`
        GPS second of the analysis

    about : `str`, optional
        relative path to the `about` page for this analysis, default: None

    Returns
    -------
    brand : `~MarkupPy.markup.page`
        the navbar brand `page` object

    class_ : `str`
        object class of the navbar
    """
    # navbar brand
    brand = markup.oneliner.div(
        ' '.join([ifo, name]),
        class_='navbar-brand border border-white rounded',
    )
    # IFO links
    page = markup.page()
    page.ul(class_='nav navbar-nav')
    page.li(class_='nav-item dropdown')
    page.a('Links', class_='nav-link dropdown-toggle',
           href='#', role='button', **{'data-toggle': 'dropdown'})
    page.div(class_='dropdown-menu dropdown-menu-right shadow')
    if about is not None:
        page.h6('Internal', class_='dropdown-header')
        page.a('About this page', href=about, class_='dropdown-item')
        page.div('', class_='dropdown-divider')
    page.h6('External', class_='dropdown-header')
    for name, url in OBSERVATORY_MAP[ifo]['links'].items():
        if 'Summary' in name:
            day = from_gps(gps).strftime(r"%Y%m%d")
            url = '/'.join([url, day])
        page.a(name, href=url, class_='dropdown-item', target='_blank')
    page.div.close()  # dropdown-menu
    page.li.close()  # nav-link dropdown-toggle
    page.ul.close()  # nav navbar-nav
    class_ = ('navbar fixed-top navbar-expand-md navbar-{} '
              'shadow-sm').format(ifo.lower())
    return ((brand, page()), class_)


def about_this_page(config, packagelist=True, prog=None):
    """Write a blurb documenting how a page was generated, including the
    command-line arguments and configuration files used

    Parameters
    ----------
    config : `str`, `list`, optional
        the absolute path(s) to one or a number of INI files used in this
        process

    packagelist : `bool`, optional
        boolean switch to include (`True`) or exclude (`False`) a
        comprehensive list of system packages

    prog : `str`, optional
        name of the program which produced this page, defaults to
        the script run on the command-line

    Returns
    -------
    page : :class:`~MarkupPy.markup.page`
        the HTML page to be inserted into the #main <div>.
    """
    page = markup.page()
    page.div(class_='row')
    page.div(class_='col-md-12')
    page.h2('On the command-line')
    page.add(get_command_line(prog=prog))
    # render config file(s)
    page.h2('Configuration files')
    page.p('The following INI-format configuration file(s) were passed '
           'on the comand-line and are reproduced here in full:')
    if isinstance(config, str):
        with open(config, 'r') as fobj:
            contents = fobj.read()
        page.add(render_code(contents, 'ini'))
    elif isinstance(config, list):
        page.div(id_='accordion')
        for i, cpfile in enumerate(config):
            page.div(class_='card mb-1 shadow-sm')
            page.div(class_='card-header')
            page.a(
                os.path.basename(cpfile),
                class_='collapsed card-link cis-link',
                href='#file%d' % i,
                **{'data-toggle': 'collapse'}
            )
            page.div.close()  # card-header
            page.div(id_='file%d' % i, class_='collapse',
                     **{'data-parent': '#accordion'})
            page.div(class_='card-body')
            with open(cpfile, 'r') as fobj:
                contents = fobj.read()
            page.add(render_code(contents, 'ini'))
            page.div.close()  # card-body
            page.div.close()  # collapse
            page.div.close()  # card
        page.div.close()  # accordion
    # render package list
    if packagelist:
        page.add(package_table())
    page.div.close()  # col-md-12
    page.div.close()  # row
    return page()


def render_code(code, language):
    """Render a block of code with syntax highlighting

    Parameters
    ----------
    code : `str`
        a raw block of source code

    language : `str`
        language the code is written in, e.g. `'python'`

    Returns
    -------
    code : `~MarkupPy.markup.page`
        fully rendered command-line call
    """
    lexer = get_lexer_by_name(language, stripall=True)
    return highlight(code, lexer, FORMATTER)


def get_command_line(language='bash', about=True, prog=None):
    """Render the command-line invocation used to generate a page

    Parameters
    ----------
    language : `str`, optional
        type of environment the code is run in, default: `'bash'`

    about : `bool`, optional
        whether this markup is for an 'about' page, default: `True`

    prog : `str`, optional
        name of the program which produced this page, defaults to
        the script run on the command-line

    Returns
    -------
    page : `~MarkupPy.markup.page`
        fully rendered command-line arguments
    """
    prog = prog or os.path.basename(sys.argv[0])
    page = markup.page()
    if about:
        page.p('This page was generated with the following command-line call:')
    args = ' '.join(sys.argv[1:])
    cmdline = ' '.join(['$', prog, args])
    page.add(render_code(cmdline.replace(' --html-only', ''), language))
    if about:
        page.p('The install path used was <code>{}</code>.'.format(sys.prefix))
    return page()


def html_link(href, txt, target="_blank", **params):
    """Write an HTML <a> tag

    Parameters
    ----------
    href : `str`
        the URL to point to

    txt : `str`
        the text for the link

    target : `str`, optional
        the ``target`` of this link

    **params
        other HTML parameters for the ``<a>`` tag

    Returns
    -------
    html : `str`
    """
    if target is not None:
        params.setdefault('target', target)
    return markup.oneliner.a(txt, href=href, **params)


def cis_link(channel, **params):
    """Write a channel name as a link to the Channel Information System

    Parameters
    ----------
    channel : `str`
        the name of the channel to link

    **params
        other HTML parmeters for the ``<a>`` tag

    Returns
    -------
    html : `str`
    """
    kwargs = {
        'title': "CIS entry for %s" % channel,
        'class_': 'cis-link',
    }
    kwargs.update(params)
    return html_link("https://cis.ligo.org/channel/byname/%s" % channel,
                     channel, **kwargs)


def fancybox_img(img, linkparams=dict(), lazy=False, **params):
    """Return the markup to embed an <img> in HTML

    Parameters
    ----------
    img : `FancyPlot`
        a `FancyPlot` object containing the path of the image to embed
        and its caption to be displayed

    linkparams : `dict`, optional
        the HTML attributes for the ``<a>`` tag

    lazy : `bool`, optional
        whether to lazy-load the image, default: False

    **params
        the HTML attributes for the ``<img>`` tag

    Returns
    -------
    page : `~MarkupPy.markup.page`
        the markup object containing fancyplot HTML
    """
    page = markup.page()
    aparams = {
        'title': img.caption,
        'class_': 'fancybox',
        'target': '_blank',
        'data-caption': img.caption,
        'data-fancybox': 'gallery',
        'data-fancybox-group': 'images',
    }
    aparams.update(linkparams)
    img = str(img)
    substrings = os.path.basename(img).split('-')
    channel = '%s-%s' % tuple(substrings[:2])
    duration = substrings[-1].split('.')[0]
    page.a(href=img, id_='a_%s_%s' % (channel, duration), **aparams)
    src_attr = lazy and 'data-src' or 'src'
    imgparams = {
        'alt': os.path.basename(img),
        'class_': lazy and 'img-fluid w-100 lazy' or 'img-fluid w-100',
        src_attr: img.replace('.svg', '.png'),
    }
    imgparams.update(params)
    page.img(id_='img_%s_%s' % (channel, duration), **imgparams)
    page.a.close()
    return page()


def scaffold_plots(plots, nperrow=3, lazy=True):
    """Embed a `list` of images in a bootstrap scaffold

    Parameters
    ----------
    plot : `list` of `FancyPlot`
        the list of image paths to embed

    nperrow : `int`
        the number of images to place in a row (on a desktop screen)

    lazy : `bool`, optional
        whether to lazy-load images, default: True

    Returns
    -------
    page : `~MarkupPy.markup.page`
        the markup object containing the scaffolded HTML
    """
    page = markup.page()
    x = int(12//nperrow)
    # scaffold plots
    for i, p in enumerate(plots):
        if i % nperrow == 0:
            page.div(class_='row scaffold')
        page.div(class_='col-sm-%d' % x)
        page.add(fancybox_img(p, lazy=lazy))
        page.div.close()  # col
        if i % nperrow == nperrow - 1:
            page.div.close()  # row
    if i % nperrow < nperrow-1:
        page.div.close()  # row
    return page()


def download_btn(content, label='Download summary',
                 btndiv='dropdown float-right d-none d-lg-block',
                 btnclass='btn btn-outline-secondary dropdown-toggle'):
    """Toggle download options with a Bootstrap button

    Parameters
    ----------
    content : `list` of `tuple` of `str`
        collection of `(title, link)` pairs to list

    label : `str`, optional
        text appearing inside the button, default: ``Download summary``

    id_ : `str`, optional
        unique HTML identifier for this button

    btndiv : `str`, optional
        class name of the enclosing ``<div>``,
        default: ``dropdown float-right d-none d-lg-block``

    btnclass : `str`, optional
        class name of the Bootstrap button object,
        default: ``btn btn-secondary dropdown-toggle``

    Returns
    -------
    page : `~MarkupPy.markup.page`
        fully rendered download button
    """
    page = markup.page()
    page.div(class_=btndiv)
    page.button(label, type='button', class_=btnclass,
                **{'data-toggle': 'dropdown',
                   'aria-expanded': 'false',
                   'aria-haspopup': 'true'})
    page.div(class_='dropdown-menu dropdown-menu-right shadow')
    for item in content:
        if len(item) == 2:
            text, href = item
            download = href
        else:
            text, href, download = item
        page.a(text, href=href, download=download,
               class_='dropdown-item')
    page.div.close()  # dropdown-menu
    page.div.close()  # btndiv
    return page()


def parameter_table(content=[], start=None, end=None, flag=None,
                    id_='parameters', tableclass='table table-sm table-hover'):
    """Render an informative section with run parameters in HTML

    Parameters
    ----------
    content: `list` of `tuple` of `str`
        collection of parameters to list

    start : `float`
        GPS start time of the analysis

    end : `float`
        GPS end time of the analysis

    flag : `str`, optional
        name of a data-quality state flag required for this analysis

    section : `str`, optional
        text to label the section header (``<h2>``), default: ``Parameters``

    id_ : `str`, optional
        unique HTML identifier for this section, default: ``parameters``

    tableclass : `str`, optional
        the ``class`` for the summary ``<table>``, defaults to a responsive
        Bootstrap table

    Returns
    -------
    page : `~MarkupPy.markup.page`
        fully rendered table of parameters
    """
    # front-load time and flag info
    common = [
        ('Start time (UTC)', '{0} ({1})'.format(from_gps(start), start)),
        ('End time (UTC)', '{0} ({1})'.format(from_gps(end), end)),
    ]
    if flag is not None:
        common.append(('State flag', markup.oneliner.code(flag)))
    content = common + content + [
        ('System prefix', markup.oneliner.code(sys.prefix))]
    # initialize page
    page = markup.page()
    page.table(class_=tableclass)
    # table body
    page.tbody()
    for row in content:
        col1, col2 = row
        page.tr()
        page.th(col1, scope='row')
        page.td(col2)
        page.tr.close()
    page.tbody.close()
    # close table and write command-line
    page.table.close()
    return page()


def alert(text, context='info', dismiss=True):
    """Enclose text within a bootstrap dialog box

    Parameters
    ----------
    text : `str` or `list` of `str`
        text to enclose within the box, if a `list` then each item will
        be rendered in a separate `<p>` tag

    context : `str`, optional
        Bootstrap context type, default: ``info``

    dismiss : `bool`, optional
        boolean switch to enable (`True`) or disable (`False`) a dismiss
        button, default: `True`

    Returns
    -------
    page : `~MarkupPy.markup.page`
        the rendered dialog box object
    """
    page = markup.page()
    class_ = ('alert alert-%s alert-dismissible fade show text-justify '
              'shadow-sm' % context if dismiss else
              'alert alert-%s text-justify shadow-sm' % context)
    page.div(class_=class_)
    if dismiss:  # add close button
        page.button(type_="button", class_="close",
                    **{'data-dismiss': 'alert', 'aria-label': 'Close'})
        page.span('&times;', **{'aria-hidden': "true"})
        page.button.close()
    if isinstance(text, (list, tuple)):
        for msg in text:
            page.p(str(msg))
    else:
        page.add(str(text))
    page.div.close()
    return page()


def table(headers, data, caption=None, separator='', id=None, **class_):
    """Write a <table> with one row of headers and many rows of data

    Parameters
    ----------
    headers : `list`
        list of column header names

    data : `list` of `lists`
        list of column data lists, for ``m`` rows and ``n`` columns, this
        should have dimensions ``m x n``

    caption : `str`, optional
        content for this table's `<caption>`

    **class_
        class attribute declarations for each tag used in the table,
        any of `table`, `thead`, `tbody`, `tr`, `th`, `td`, `caption`

    Returns
    -------
    table : `~MarkupPy.markup.page`
        a formatted HTML page object containing the `<table>`
    """
    class_.setdefault('table', 'table table-sm table-hover')
    # unwrap class declarations (so we don't get empty class attributes)
    kwargs = {}
    for tag in ['table', 'thead', 'tbody', 'tr', 'th', 'td', 'caption']:
        try:
            kwargs[tag] = {'class_': class_.pop(tag)}
        except KeyError:
            kwargs[tag] = {}
    # create table and add caption
    page = markup.page(separator=separator)
    if id is not None:
        kwargs['table']['id_'] = id
    page.table(**kwargs['table'])
    if caption:
        page.caption(caption, **kwargs['caption'])
    # write headers
    page.thead(**kwargs['thead'])
    page.tr(**kwargs['tr'])
    for th in headers:
        page.th(th, scope='col', **kwargs['th'])
    page.tr.close()
    page.thead.close()
    # write body
    page.tbody(**kwargs['tbody'])
    for row in data:
        page.tr(**kwargs['tr'])
        for td in row:
            page.td(td, **kwargs['td'])
        page.tr.close()
    page.tbody.close()
    page.table.close()
    # add export button
    if id:
        page.button(
            'Export to CSV', class_='btn btn-outline-secondary btn-table mt-2',
            **{'data-table-id': id, 'data-filename': '%s.csv' % id})
    return page()


def write_flag_html(flag, span=None, id=0, parent='accordion',
                    context='warning', title=None, plotdir=None,
                    plot_func=plot_segments, **kwargs):
    """Write HTML for data-quality flags

    Parameters
    ----------
    flag : `~gwpy.segments.DataQualityFlag`
        the flag object to represent

    span : `~gwpy.sements.Segment`, optional
        GPS start and stop time of flag duration

    id : `int`, optional
        unique identifier for the flag, default: 0

    parent : `str`, optional
        HTML identifier of the parent object, default: ``accordion``

    context : `str`, optional
        Bootstrap context type, default: ``warning``

    title : `str`, optional
        title of the flag, defaults to `flag.name`

    plotdir : `str`, optional
        path to directory containing plots, required if plotting segments

    plot_func : `func`, optional
        function used to plot segments,
        default: `~gwdetchar.plot.plot_segments`

    **kwargs : `dict`, optional
        additional keyword arguments to `plot_func`

    Returns
    -------
    page : `~MarkupPy.markup.page`
        fully rendered HTML with a card object for the flag
    """
    page = markup.page()
    page.div(class_='card border-%s mb-1 shadow-sm' % context)
    page.div(class_=_get_card_header(context))
    title = title or flag.name
    page.a(title, class_='collapsed card-link cis-link', href='#flag%s' % id,
           **{'data-toggle': 'collapse'})
    page.div.close()  # card-header
    page.div(id_='flag%s' % id, class_='collapse',
             **{'data-parent': '#' + parent})
    page.div(class_='card-body')
    # render segment plot
    if (plotdir is not None) and (plot_func is not None):
        flagr = flag.name.replace('-', '_').replace(':', '-', 1)
        png = os.path.join(
            plotdir, '%s-%d-%d.png' % (flagr, span[0], abs(span)))
        plot = plot_func(flag, span, **kwargs)
        plot.save(png)
        plot.close()
        # set up fancybox
        img = FancyPlot(
            img=png, caption='Known (small) and active (large) analysis '
                             'segments for {}'.format(title))
        page.add(scaffold_plots([img], nperrow=1, lazy=False))
    # write segments
    segs = StringIO()
    try:
        flag.active.write(segs, format='segwizard',
                          coltype=type(flag.active[0][0]))
    except IndexError:
        page.p('No segments were found.')
    else:
        page.pre(segs.getvalue())
    page.div.close()  # card-body
    page.div.close()  # collapse
    page.div.close()  # card
    return page()


def scaffold_omega_scans(times, channel, plot_durations=[1, 4, 16],
                         scandir=os.path.curdir):
    """Preview a batch of omega scans in HTML

    Parameters
    ----------
    times : `list` of `float`
        a list of GPS times to scan

    channel : `str`
        name of the channel being scanned, e.g. ``H1:GDS-CALIB_STRAIN``

    plot_durations : `list` of `int`
        list of plot durations in seconds, default: `[1, 4, 16]`

    scandir : `str`
        path to the directory containing omega scans, default: `.`

    Returns
    -------
    page : `~MarkupPy.markup.page`
        rendered scaffold of omega scan plots
    """
    ifo = channel[:2].lower()
    page = markup.page()
    page.div(class_='card card-%s' % ifo)
    page.div(class_='card-header pb-0')
    page.h5(cis_link(channel), class_='card-title')
    page.div.close()  # card-header
    page.div(class_='card-body')
    page.ul(class_='list-group')
    for t in times:
        page.li(class_='list-group-item')
        page.div(class_='container')
        page.div(class_='row')
        page.h6()
        page.a(t, href='{0}/{1}'.format(scandir, t), class_='text-dark')
        page.h6.close()
        page.div.close()  # row
        chanstr = channel.replace('-', '_').replace(':', '-')
        plots = [
            '{}/{}/plots/{}-qscan_whitened-{}.png'.format(
                scandir, t, chanstr, dur) for dur in plot_durations]
        page.add(scaffold_plots(
            [FancyPlot(plot) for plot in plots],
            nperrow=3))
        page.div.close()  # container
        page.li.close()  # list-group-item
    page.ul.close()  # list-group
    page.div.close()  # card-body
    page.div.close()  # card
    return page()


def write_footer(about=None, link=None, issues=None, external=None):
    """Write a <footer> for a bootstrap page

    Parameters
    ----------
    about : `str`, optional
        path of about page to link

    link : `None` or `tuple`, optional
        tuple of package name, HTML link to source code, and host name
        (e.g. GitHub)

    issues : `str`, optional
        HTML link to issue report page

    external : `str`, optional
        additional footer link to an external page

    Returns
    -------
    page : `~MarkupPy.markup.page`
        the markup object containing the footer HTML
    """
    page = markup.page()
    page.twotags.append('footer')
    markup.element('footer', case=page.case, parent=page)(class_='footer')
    page.div(class_='container')
    if link is None:
        version = get_versions()['version']
        commit = get_versions()['full-revisionid']
        package = 'gwdetchar-%s' % version
        source = 'https://github.com/gwdetchar/gwdetchar/tree/%s' % commit
        host = 'GitHub'
    elif isinstance(link, (tuple, list)):
        package, source, host = link
    else:
        raise ValueError("'link' argument must be either None or a tuple of "
                         "package name, URL, and host name")
    # format various links
    page.div(class_='row')
    page.div(class_='col-sm-3 icon-bar')
    page.a(markup.oneliner.i('', class_='fas fa-code'), href=source,
           title='View {0} on {1}'.format(package, host), target='_blank')
    page.a(markup.oneliner.i('', class_='fas fa-ticket-alt'),
           href=issues or 'https://github.com/gwdetchar/gwdetchar/issues',
           title='Open an issue ticket', target='_blank')
    if about is not None:
        page.a(markup.oneliner.i('', class_='fas fa-info-circle'),
               href=about, title='How was this page generated?')
    if external is not None:
        page.a(markup.oneliner.i('', class_='fas fa-external-link-alt'),
               href=external, title="View this page's external source")
    page.a(markup.oneliner.i('', class_='fas fa-heartbeat'),
           href='https://apod.nasa.gov/apod/astropix.html',
           title='Take a break from science', target='_blank')
    page.div.close()  # col-sm-3 icon-bar
    # print timestamp
    page.div(class_='col-sm-6')
    now = datetime.datetime.now()
    tz = reference.LocalTimezone().tzname(now)
    date = now.strftime('%H:%M {} on %d %B %Y'.format(tz))
    page.p('Created by {0} at {1}'.format(getuser(), date))
    page.div.close()  # col-sm-6
    page.div.close()  # row
    page.div.close()  # container
    markup.element('footer', case=page.case, parent=page).close()
    return page()


def close_page(page, target, **kwargs):
    """Close an HTML document with markup then write to disk

    This method writes the closing markup to complement the opening
    written by `init_page`, something like:

    .. code:: html

       </div>
       <footer>
           <!-- some stuff -->
       </footer>
       </body>
       </html>

    Parameters
    ----------
    page : `markup.page`
        the markup object to close

    target : `str`
        the output filename for HTML
    """
    page.div.close()  # container
    page.add(write_footer(**kwargs))
    if not page._full:
        page.body.close()
        page.html.close()
    with open(target, 'w') as f:
        f.write(page())
    return page()


def package_list():
    """Get the list of packages installed alongside this one

    Returns a `list` of `dict`
    """
    prefix = sys.prefix
    if (Path(prefix) / "conda-meta").is_dir():
        raw = subprocess.check_output(
            ["conda", "list",
             "--prefix", prefix,
             "--json"],
        )
    else:
        raw = subprocess.check_output(
            [sys.executable,
             "-m", "pip",
             "list", "installed",
             "--format", "json"],
        )
    if isinstance(raw, bytes):
        raw = raw.decode('utf-8')
    return json.loads(raw)


def package_table(
        h2="Environment",
        class_='table table-sm table-hover table-responsive',
        caption="Table of packages installed in the production environment",
        id_="package-table",
):
    """Write a table listing packages installed in the current environment

    Parameters
    ----------
    h2 : `str`, `None`, optional
        the header for the HTML section

    caption : `str`, `None`, optional
        the `<caption>` for the package table

    Returns
    -------
    html : `str`
        an HTML table
    """
    # get package list and inspect columns
    pkgs = package_list()
    if "build_string" in pkgs[0]:  # conda list
        cols = ("name", "version", "channel", "build_string")
    else:  # pip list installed
        cols = ("name", "version")
    # create page and write <table>
    page = markup.page(separator="")
    if h2 is not None:
        page.h2(h2, class_='mt-4')
    headers = [head.title() for head in cols]
    data = [[pkg[col.lower()] for col in cols]
            for pkg in sorted(pkgs, key=itemgetter("name"))]
    page.add(table(headers, data, caption=caption, id=id_, table=class_))
    return page()



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwdetchar.scattering.plot`
"""

import os
import numpy
import shutil

from gwpy.timeseries import TimeSeries

from matplotlib import use
use('Agg')

# backend-dependent import
from .. import plot  # noqa: E402

__author__ = 'Alex Urban <alexander.urban@ligo.org>'


# global test objects

TWOPI = 2 * numpy.pi
TIMES = numpy.arange(0, 16384 * 64)
NOISE = TimeSeries(
    numpy.random.normal(loc=1, scale=.5, size=16384 * 64),
    sample_rate=16384, epoch=-32).zpk([], [0], 1)
FRINGE = TimeSeries(
    numpy.cos(TWOPI * TIMES), sample_rate=16384, epoch=-32)
DATA = NOISE.inject(FRINGE)
QSPECGRAM = DATA.q_transform(logf=True, method="median")


# -- make sure plots run end-to-end -------------------------------------------

def test_spectral_comparison(tmpdir):
    outdir = str(tmpdir)
    plot1 = os.path.join(outdir, 'test1.png')
    plot2 = os.path.join(outdir, 'test2.png')
    # test plotting
    plot.spectral_comparison(0, QSPECGRAM, FRINGE, plot1)
    plot.spectral_overlay(0, QSPECGRAM, FRINGE, plot2)
    # clean up
    shutil.rmtree(outdir, ignore_errors=True)



# -*- coding: utf-8 -*-
# Copyright (C) Joshua Smith (2016-)
#
# This file is part of the GW DetChar python package.
#
# gwdetchar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# gwdetchar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Test suite for `gwdetchar.condor`
"""

import os
import tempfile

import pytest

from .. import (condor, const)

LAST_CONDOR_EPOCH = list(filter(
    condor.OBS_RUN_REGEX.match,
    list(zip(*sorted(const.EPOCH.items(), key=lambda x: x[1].start)))[0],
))[-1]


@pytest.mark.parametrize('gps, epoch', [
    (0, 'S6'),  # earlier than first run, pin to first
    (1126259462, 'O1'),  # in a run
    (2e50, LAST_CONDOR_EPOCH),  # beyond last run, pin to latest
])
def test_accounting_epoch(gps, epoch):
    assert condor.accounting_epoch(gps) == epoch


@pytest.fixture
def tagfile():
    name = tempfile.mktemp()
    with open(name, 'w') as tmp:
        tmp.write("* tag1 found\n* tag2 found\n* tag3 found\n")
    try:
        yield name
    finally:
        os.remove(name)


@pytest.mark.parametrize('tag, result', [
    ('tag1', True),
    ('something else', False),
])
def test_validate_accounting_tag(tagfile, tag, result):
    assert condor.validate_accounting_tag(tag, path=tagfile) == result
    if result:
        assert condor.is_valid(tag, path=tagfile) == result
    else:
        with pytest.raises(ValueError):
            condor.is_valid(tag, path=tagfile)



# coding=utf-8
# Copyright (C) LIGO Scientific Collaboration (2015-)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Find times when an input signal crosses a particular threshold
"""

import gwdatafind
import os
import sys

from astropy.table import vstack as vstack_tables

from gwpy.io.cache import (cache_segments, sieve as sieve_cache)
from gwpy.segments import (DataQualityFlag, Segment, SegmentList)
from gwpy.table import EventTable

from . import (const, cli)
from .daq import find_crossings
from .io.datafind import get_data
from .utils import table_from_times

__author__ = 'TJ Massinger <thomas.massinger@ligo.org>'
__credits__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

# set up logger
PROG = ('python -m gwdetchar.mct' if sys.argv[0].endswith('.py')
        else os.path.basename(sys.argv[0]))
LOGGER = cli.logger(name=PROG.split('python -m ').pop())


# -- parse command-line -------------------------------------------------------

def create_parser():
    """Create a command-line parser for this entry point
    """
    # initialize argument parser
    parser = cli.create_parser(
        prog=PROG,
        description=__doc__,
    )

    # positional arguments
    cli.add_gps_start_stop_arguments(parser)
    cli.add_ifo_option(parser)
    cli.add_nproc_option(parser)
    cli.add_frametype_option(
        parser,
        required=const.IFO is None,
        default=const.IFO is not None and '%s_R' % const.IFO,
    )

    # optional arguments
    parser.add_argument(
        '-a',
        '--state-flag',
        metavar='FLAG',
        help='restrict search to times when FLAG was active',
    )
    parser.add_argument(
        '-o',
        '--output-path',
        help='path to output HDF5 file, name will be '
             'automatically generated based on IFO and GPS times',
    )
    parser.add_argument(
        '-t',
        '--threshold',
        type=float,
        nargs='+',
        default=[0., 2.**16, -2.**16],
        help='threshold for marking input data crossings',
    )
    parser.add_argument(
        '-c',
        '--channel',
        required=True,
        type=str,
        help='channel to read for input data',
    )
    parser.add_argument(
        '-r',
        '--rate-thresh',
        default=16.,
        type=float,
        help='if the trigger rate (Hz) is above this value for a given '
             'segment, crossings for that segment will not be recorded',
    )

    # return the argument parser
    return parser


# -- main code block ----------------------------------------------------------

def main(args=None):
    """Run the zero-crossing counter tool
    """
    parser = create_parser()
    args = parser.parse_args(args=args)

    span = Segment(args.gpsstart, args.gpsend)
    LOGGER.info('-- Processing channel %s over span %d - %d'
                % (args.channel, args.gpsstart, args.gpsend))

    if args.state_flag:
        state = DataQualityFlag.query(
            args.state_flag,
            int(args.gpsstart),
            int(args.gpsend),
            url=const.DEFAULT_SEGMENT_SERVER,
        )
        statea = state.active
    else:
        statea = SegmentList([span])

    duration = abs(span)

    # initialize output files for each threshold and store them in a dict
    outfiles = {}
    for thresh in args.threshold:
        outfiles[str(thresh)] = (
            os.path.join(
                args.output_path,
                '%s_%s_DAC-%d-%d.h5'
                % (args.channel.replace('-', '_').replace(':', '-'),
                   str(int(thresh)).replace('-', 'n'),
                   int(args.gpsstart), duration)))

    # get frame cache
    cache = gwdatafind.find_urls(args.ifo[0], args.frametype,
                                 int(args.gpsstart), int(args.gpsend))

    cachesegs = statea & cache_segments(cache)

    if not os.path.exists(args.output_path):
        os.makedirs(args.output_path)

    # initialize a ligolw table for each threshold and store them in a dict
    names = ("time", "frequency", "snr")
    dtypes = ("f8",) * len(names)
    tables = {}
    for thresh in args.threshold:
        tables[str(thresh)] = EventTable(
            names=names,
            dtype=dtypes,
            meta={"channel": args.channel},
        )

    # for each science segment, read in the data from frames, check for
    # threshold crossings, and if the rate of crossings is less than
    # rate_thresh, write to a sngl_burst table
    for seg in cachesegs:
        LOGGER.debug("Processing {}:".format(seg))
        c = sieve_cache(cache, segment=seg)
        if not c:
            LOGGER.warning("    No {} data files for this segment, "
                           "skipping".format(args.frametype))
            continue
        data = get_data(args.channel, seg[0], seg[1], nproc=args.nproc,
                        source=c, verbose="Reading data:".rjust(30))
        for thresh in args.threshold:
            times = find_crossings(data, thresh)
            rate = float(times.size)/abs(seg) if times.size else 0
            LOGGER.info("    Found {0} crossings of {1}, rate: {2} Hz".format(
                times.size,
                thresh,
                rate,
            ))
            if times.size and rate < args.rate_thresh:
                existing = tables[str(thresh)]
                tables[str(thresh)] = vstack_tables(
                    (existing,
                     table_from_times(times, snr=10., frequency=100.,
                                      names=existing.colnames),
                     ),
                    join_type="exact",
                )

    n = max(map(len, tables.values()))
    for thresh, outfile in outfiles.items():
        tables[thresh].write(
            outfile,
            path="triggers",
            format="hdf5",
            overwrite=True,
        )
        LOGGER.info("{0} events written to {1}".format(
            str(len(tables[thresh])).rjust(len(str(n))),
            outfile,
        ))


# -- run from command-line ----------------------------------------------------

if __name__ == "__main__":
    main()



# coding=utf-8
# Copyright (C) Duncan Macleod (2015)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Data discovery utilities
"""

import re
import warnings
from urllib.error import HTTPError
from json.decoder import JSONDecodeError

from ..const import DEFAULT_SEGMENT_SERVER

from gwpy.io import gwf as io_gwf
from gwpy.io import datafind as io_datafind
from gwpy.segments import (Segment, DataQualityFlag)
from gwpy.timeseries import (TimeSeries, TimeSeriesDict)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Alex Urban <alexander.urban@ligo.org>'


# -- utilities ----------------------------------------------------------------

def check_flag(flag, gpstime, duration, pad):
    """Check that a state flag is active during an entire analysis segment

    Parameters
    ----------
    flag : `str`
        state flag to check

    gpstime : `float`
        GPS time of required data

    duration : `float`
        duration (in seconds) of required data

    pad : `float`
        amount of extra data to read in at the start and end for filtering

    Returns
    -------
    check : `bool`
        Boolean switch to pass (`True`) or fail (`False`) depending on whether
        the given flag is active
    """
    # set GPS start and end time
    start = gpstime - duration/2. - pad
    end = gpstime + duration/2. + pad
    seg = Segment(start, end)
    # query for state segments
    active = DataQualityFlag.query(flag, start, end,
                                   url=DEFAULT_SEGMENT_SERVER).active
    # check that state flag is active during the entire analysis
    if (not active.intersects_segment(seg)) or (abs(active[0]) < abs(seg)):
        return False
    return True


def remove_missing_channels(channels, gwfcache):
    """Find and remove channels from a given list that are not available in
    a given cache of frame files

    Parameters
    ----------
    channels : `list` of `str`
        list of requested channels

    gwfcache : `list` of `str`
        list of paths to .gwf files

    Returns
    -------
    keep : `list` of `str`
        list of common channels found in the first and last files in the
        cache

    Notes
    -----
    As a shorthand, this utility checks `channels` against only the first
    and last frame files in `gwfcache`. This saves time and memory by not
    loading tables of contents for large numbers of very long data files.

    For every channel requested that is not available in `gwfcache`, a
    `UserWarning` will be raised.

    See Also
    --------
    gwpy.io.gwf.iter_channel_names
        for the utility used to identify frame contents
    """
    # get available channels from the first and last frame file
    available = set(io_gwf.iter_channel_names(gwfcache[0]))
    if len(gwfcache) > 1:
        available.intersection_update(io_gwf.iter_channel_names(gwfcache[-1]))
    # work out which channels to keep, and which to reject
    channels = set(channels)
    keep = channels & available
    reject = channels - keep
    for channel in reject:
        warnings.warn(
            '{} is being removed because it was not available in all '
            'requested files'.format(channel), UserWarning)
    return list(keep)


def get_data(channel, start, end, frametype=None, source=None,
             nproc=1, verbose=False, **kwargs):
    """Retrieve data for given channels within a certain time range

    Parameters
    ----------
    channel : `str` or `list`
        either a single channel name, or a list of channel names

    start : `float`
        GPS start time of requested data

    end : `float`
        GPS end time of requested data

    frametype : `str`, optional
        name of frametype in which channel(s) are stored, default: `None`

    source : `str`, `list`, optional
        path(s) of a LAL-format cache file or individual data file,
        default: `None`

    nproc : `int`, optional
        number of parallel processes to use, uses serial process by default

    verbose : `bool`, optional
        print verbose output about NDS progress, default: False

    **kwargs : `dict`, optional
        additional keyword arguments to `~gwpy.timeseries.TimeSeries.read`
        or `~gwpy.timeseries.TimeSeries.get`

    Returns
    -------
    data : `~gwpy.timeseries.TimeSeries` or `~gwpy.timeseries.TimeSeriesDict`
        collection of data for the requested channels in the requested time
        range

    Notes
    -----
    If `channel` is a `str`, then a `TimeSeries` object will be returned, else
    the result is a `TimeSeriesDict`.

    The `frametype` argument should be used to read from archived frame files,
    while `source` should be used to read from a local cache or specific data
    file. If either fails, or if neither is passed, this function will attempt
    to get data over an NDS server.

    If `frametype` is used to read from the archive, any channels missing
    from the first or last frame file in the requested time range will be
    ignored.

    See Also
    --------
    remove_missing_channels
        a utility that removes channels missing from the frame archive
    gwpy.timeseries.TimeSeries.get
        the underlying method to read data over an NDS server
    gwpy.timeseries.TimeSeries.read
        the underlying method to read data from local files
    """
    # get TimeSeries class
    if isinstance(channel, (list, tuple)):
        series_class = TimeSeriesDict
    else:
        series_class = TimeSeries

    if frametype is not None:
        try:  # locate frame files
            ifo = re.search('[A-Z]1', frametype).group(0)
            obs = ifo[0]
            source = io_datafind.find_urls(obs, frametype, start, end)
        except AttributeError:
            raise AttributeError(
                'Could not determine observatory from frametype')
        except (HTTPError, JSONDecodeError):  # frame files not found
            pass
    if source and (isinstance(source, list) and
                   isinstance(channel, (list, tuple))):
        channel = remove_missing_channels(channel, source)
    if source:  # read from frame files
        return series_class.read(
            source, channel, start=start, end=end, nproc=nproc,
            verbose=verbose, **kwargs)

    # read single channel from NDS
    if not isinstance(channel, (list, tuple)):
        return series_class.get(
            channel, start, end, verbose=verbose, **kwargs)

    # if all else fails, process channels in groups of 60
    data = series_class()
    for group in [channel[i:i + 60] for i in range(0, len(channel), 60)]:
        data.append(series_class.get(
            group, start, end, verbose=verbose, **kwargs))
    return data



# coding=utf-8
# Copyright (C) LIGO Scientific Collaboration (2015-)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Find ADC/DAC overflows associated with a particular front-end model
"""

import gwdatafind
import h5py
import os
import sys
import tqdm
import warnings

from gwpy.io.cache import cache_segments
from gwpy.segments import (DataQualityFlag, DataQualityDict,
                           Segment, SegmentList)

from . import (cds, cli, const, daq)
from .io import (html as htmlio)
from .io.datafind import get_data
from .utils import table_from_segments

from matplotlib import use
use('Agg')

__author__ = 'TJ Massinger <thomas.massinger@ligo.org>'
__credits__ = ('Duncan Macleod <duncan.macleod@ligo.org> '
               'Alex Urban <alexander.urban@ligo.org>')

SITE_MAP = {
    'H1': 'LIGO-Hanford',
    'L1': 'LIGO-Livingston',
    'K1': 'KAGRA',
    'V1': 'Virgo',
}

# set up logger
PROG = ('python -m gwdetchar.overflow' if sys.argv[0].endswith('.py')
        else os.path.basename(sys.argv[0]))
LOGGER = cli.logger(name=PROG.split('python -m ').pop())


# -- parse command-line -------------------------------------------------------

def create_parser():
    """Create a command-line parser for this entry point
    """
    # initialize argument parser
    parser = cli.create_parser(
        prog=PROG,
        description=__doc__,
    )

    # required arguments
    cli.add_gps_start_stop_arguments(parser)
    cli.add_ifo_option(parser)

    # optional arguments
    parser.add_argument(
        'dcuid',
        type=int,
        nargs='+',
        help='DCUID for the relevant front-end model',
    )
    cli.add_frametype_option(
        parser,
        required=const.IFO is None,
        default=const.IFO is not None and '%s_R' % const.IFO,
    )
    cli.add_nproc_option(
        parser,)

    parser.add_argument(
        '--deep',
        action='store_true',
        default=False,
        help='perform deep scan, default: %(default)s',
    )
    parser.add_argument(
        '-a',
        '--state-flag',
        metavar='FLAG',
        help='restrict search to times when FLAG was active',
    )
    parser.add_argument(
        '--nds',
        metavar='HOST:PORT',
        help='use the given NDS host:port to access data, '
             'default: use datafind',
    )
    parser.add_argument(
        '-o',
        '--output-file',
        help='path to output data file, default name will be '
             'automatically generated based on IFO and GPS times',
    )
    parser.add_argument(
        '-I',
        '--integer-segments',
        action='store_true',
        default=False,
        help='pad all overflow segments to integer '
             'boundaries (default: %(default)s)',
    )
    parser.add_argument(
        '-p',
        '--segment-pad',
        type=float,
        default=0.1,
        help='minimum padding (one-sided) for output segments when '
             'using --output-format [segments|integer-segments]',
    )
    parser.add_argument(
        '-s',
        '--segment-end-pad',
        type=float,
        default=1.0,
        help='amount of time to remove from the end of each analysis segment',
    )
    parser.add_argument(
        '-m',
        '--html',
        type=os.path.abspath,
        help='path to write html output',
    )
    parser.add_argument(
        '-v',
        '--plot',
        action='store_true',
        default=None,
        help='make plots of all overflows, default: %(default)s',
    )
    parser.add_argument(
        '-c',
        '--fec-map',
        help='URL of human-readable FEC map, default: infer from IFO',
    )
    parser.add_argument(
        '-u',
        '--simulink',
        help='URL of human-readable Simulink model, default: infer from IFO',
    )
    parser.add_argument(
        '-d',
        '--daqsvn',
        help='URL of the front-end data gathering configuration, '
             'default: internal LIGO DAQ subversion repository',
    )

    # return the argument parser
    return parser


# -- main code block ----------------------------------------------------------

def main(args=None):
    """Run the online Guardian node visualization tool
    """
    parser = create_parser()
    args = parser.parse_args(args=args)

    fec_map = args.fec_map
    simulink = args.simulink
    daqsvn = args.daqsvn or ('https://daqsvn.ligo-la.caltech.edu/websvn/'
                             'listing.php?repname=daq_maps')
    if args.ifo == 'H1':
        if not fec_map:
            fec_map = 'https://lhocds.ligo-wa.caltech.edu/exports/detchar/fec/'
        if not simulink:
            simulink = 'https://lhocds.ligo-wa.caltech.edu/daq/simulink/'
    if args.ifo == 'L1':
        if not fec_map:
            fec_map = 'https://llocds.ligo-la.caltech.edu/exports/detchar/fec/'
        if not simulink:
            simulink = 'https://llocds.ligo-la.caltech.edu/daq/simulink/'

    span = Segment(args.gpsstart, args.gpsend)

    # let's go
    LOGGER.info('{} Overflows {}-{}'.format(
        args.ifo, int(args.gpsstart), int(args.gpsend)))

    # get segments
    if args.state_flag:
        state = DataQualityFlag.query(args.state_flag, int(args.gpsstart),
                                      int(args.gpsend),
                                      url=const.DEFAULT_SEGMENT_SERVER)
        tmp = type(state.active)()
        for i, seg in enumerate(state.active):
            if abs(seg) < args.segment_end_pad:
                continue
            tmp.append(type(seg)(seg[0], seg[1]-args.segment_end_pad))
        state.active = tmp.coalesce()
        statea = state.active
    else:
        statea = SegmentList([span])

    if not args.output_file:
        duration = abs(span)
        args.output_file = (
            '%s-OVERFLOWS-%d-%d.h5'
            % (args.ifo, int(args.gpsstart), duration))
        LOGGER.debug("Set default output file as %s" % args.output_file)

    # set up container
    overflows = DataQualityDict()

    # prepare data access
    if args.nds:
        from gwpy.io import nds2 as io_nds2
        host, port = args.nds.rsplit(':', 1)
        ndsconnection = io_nds2.connect(host, port=int(port))
        if ndsconnection.get_protocol() == 1:
            cachesegs = SegmentList([Segment(int(args.gpsstart),
                                             int(args.gpsend))])
        else:
            cachesegs = io_nds2.get_availability(
                ['{0}:FEC-1_DAC_OVERFLOW_ACC_0_0'.format(args.ifo)],
                int(args.gpsstart), int(args.gpsend),
            )
    else:  # get frame cache
        cache = gwdatafind.find_urls(args.ifo[0], args.frametype,
                                     int(args.gpsstart), int(args.gpsend))
        cachesegs = statea & cache_segments(cache)

    flag_desc = "ADC/DAC Overflow indicated by {0}"

    # get channel and find overflows
    for dcuid in args.dcuid:
        LOGGER.info("Processing DCUID %d" % dcuid)
        channel = daq.ligo_accum_overflow_channel(dcuid, args.ifo)
        overflows[channel] = DataQualityFlag(channel, known=cachesegs)
        if args.deep:
            LOGGER.debug(" -- Getting list of overflow channels")
            try:
                channels = daq.ligo_model_overflow_channels(
                    dcuid, args.ifo, args.frametype,
                    gpstime=span[0], nds=args.nds)
            except IndexError:  # no frame found for GPS start, try GPS end
                channels = daq.ligo_model_overflow_channels(
                    dcuid, args.ifo, args.frametype, gpstime=span[-1])
            for chan in channels:  # set up flags early
                overflows[chan] = DataQualityFlag(
                    chan,
                    known=cachesegs,
                    description=flag_desc.format(chan),
                    isgood=False,
                )
            LOGGER.debug(" -- %d channels found" % len(channel))
        for seg in cachesegs:
            LOGGER.debug(" -- Processing {}-{}".format(*seg))
            if args.nds:
                read_kw = dict(connection=ndsconnection)
            else:
                read_kw = dict(source=cache, nproc=args.nproc)
            msg = "Reading ACCUM_OVERFLOW data:".rjust(30)
            data = get_data(
                channel, seg[0], seg[1],
                pad=0., verbose=msg, **read_kw
            )
            new = daq.find_overflow_segments(
                data,
                cumulative=True,
            )
            overflows[channel] += new
            LOGGER.info(" -- {} overflows found".format(len(new.active)))
            if not new.active:
                continue
            # go deep!
            for s, e in tqdm.tqdm(new.active.protract(2), unit='ovfl',
                                  desc='Going deep'.rjust(30)):
                data = get_data(channels, s, e, **read_kw)
                for ch in channels:
                    try:
                        overflows[ch] += daq.find_overflow_segments(
                            data[ch],
                            cumulative=True,
                        )
                    except KeyError:
                        warnings.warn("Skipping {}".format(ch), UserWarning)
                        continue
        LOGGER.debug(" -- Search complete")

    # write output
    LOGGER.info("Writing segments to %s" % args.output_file)
    table = table_from_segments(
        overflows,
        sngl_burst=args.output_file.endswith((".xml", ".xml.gz")),
    )
    if args.integer_segments:
        for key in overflows:
            overflows[key] = overflows[key].round()
    if args.output_file.endswith((".h5", "hdf", ".hdf5")):
        with h5py.File(args.output_file, "w") as h5f:
            table.write(h5f, path="triggers")
            overflows.write(h5f, path="segments")
    else:
        table.write(args.output_file, overwrite=True)
        overflows.write(args.output_file, overwrite=True, append=True)

    # write HTML
    if args.html:
        # get base path
        base = os.path.dirname(args.html)
        os.chdir(base)
        if args.plot:
            args.plot = os.path.curdir
        if args.output_file:
            args.output_file = os.path.relpath(
                args.output_file, os.path.dirname(args.html))
        if os.path.basename(args.html) == 'index.html':
            links = [
                '%d-%d' % (int(args.gpsstart), int(args.gpsend)),
                ('Parameters', '#parameters'),
                ('Segments', [('Overflows', '#overflows')]),
                ('Results', '#results'),
            ]
            if args.state_flag:
                links[2][1].insert(0, ('State flag', '#state-flag'))
            (brand, class_) = htmlio.get_brand(args.ifo, 'Overflows',
                                               args.gpsstart)
            navbar = htmlio.navbar(links, class_=class_, brand=brand)
            page = htmlio.new_bootstrap_page(
                title='%s Overflows | %d-%d' % (
                    args.ifo, int(args.gpsstart), int(args.gpsend)),
                navbar=navbar)
        else:
            page = htmlio.markup.page()
            page.div(class_='container')

        # -- header
        page.div(class_='pb-2 mt-3 mb-2 border-bottom')
        page.h1('%s ADC/DAC Overflows: %d-%d'
                % (args.ifo, int(args.gpsstart), int(args.gpsend)))
        page.div.close()

        # -- paramters
        content = [('DCUIDs', ' '.join(map(str, args.dcuid)))]
        if daqsvn:
            content.append((
                'FEC configuration',
                ('<a href="{0}" target="_blank" title="{1} FEC configuration">'
                 '{0}</a>').format(daqsvn, args.ifo)))
        if fec_map:
            content.append((
                'FEC map', '<a href="{0}" target="_blank" title="{1} FEC '
                           'map">{0}</a>'.format(fec_map, args.ifo)))
        if simulink:
            content.append((
                'Simulink models', '<a href="{0}" target="_blank" title="{1} '
                                   'Simulink models">{0}</a>'.format(
                                       simulink, args.ifo)))
        page.h2('Parameters', class_='mt-4 mb-4', id_='parameters')
        page.div(class_='row')
        page.div(class_='col-md-9 col-sm-12')
        page.add(htmlio.parameter_table(
            content, start=args.gpsstart, end=args.gpsend,
            flag=args.state_flag))
        page.div.close()  # col-md-9 col-sm-12

        # link to summary file
        if args.output_file:
            ext = ('HDF' if args.output_file.endswith((".h5", "hdf", ".hdf5"))
                   else 'XML')
            page.div(class_='col-md-3 col-sm-12')
            page.add(htmlio.download_btn(
                [('Segments ({})'.format(ext), args.output_file)],
                btnclass='btn btn-%s dropdown-toggle' % args.ifo.lower(),
            ))
            page.div.close()  # col-md-3 col-sm-12
        page.div.close()  # row

        # -- command-line
        page.h5('Command-line:')
        page.add(htmlio.get_command_line(about=False, prog=PROG))

        # -- segments
        page.h2('Segments', class_='mt-4', id_='segments')

        # give contextual information
        msg = ("This analysis searched for digital-to-analogue (DAC) or "
               "analogue-to-digital (ADC) conversion overflows in the {0} "
               "real-time controls system. ").format(
                   SITE_MAP.get(args.ifo, 'LIGO'))
        if args.deep:
            msg += (
                "A hierarchichal search was performed, with one cumulative "
                "overflow counter checked per front-end controller (FEC). "
                "For those models that indicated an overflow, the card- and "
                "slot-specific channels were then checked. "
            )
        msg += (
            "Consant overflow is shown as yellow, while transient overflow "
            "is shown as red. If a data-quality flag was loaded for this "
            "analysis, it will be displayed in green."
        )
        page.add(htmlio.alert(msg, context=args.ifo.lower()))
        # record state segments
        if args.state_flag:
            page.h3('State flag', class_='mt-3', id_='state-flag')
            page.div(id_='accordion1')
            page.add(htmlio.write_flag_html(
                state, span, 'state', parent='accordion1', context='success',
                plotdir=args.plot, facecolor=(0.2, 0.8, 0.2),
                edgecolor='darkgreen', known={
                    'facecolor': 'red',
                    'edgecolor': 'darkred',
                    'height': 0.4,
                }
            ))
            page.div.close()
        # record overflow segments
        if sum(abs(s.active) for s in overflows.values()):
            page.h3('Overflows', class_='mt-3', id_='overflows')
            page.div(id_='accordion2')
            for i, (c, flag) in enumerate(list(overflows.items())):
                if abs(flag.active) == 0:
                    continue
                if abs(flag.active) == abs(cachesegs):
                    context = 'warning'
                else:
                    context = 'danger'
                try:
                    channel = cds.get_real_channel(flag.name)
                except Exception:
                    title = '%s [%d]' % (flag.name, len(flag.active))
                else:
                    title = '%s (%s) [%d]' % (flag.name, channel,
                                              len(flag.active))
                page.add(htmlio.write_flag_html(
                    flag, span, i, parent='accordion2', title=title,
                    context=context, plotdir=args.plot))
            page.div.close()
        else:
            page.add(htmlio.alert('No overflows were found in this analysis',
                                  context=args.ifo.lower(), dismiss=False))

        # -- results table
        page.h2('Results summary', class_='mt-4', id_='results')
        page.table(class_='table table-striped table-hover')
        # write table header
        page.thead()
        page.tr()
        for header in ['Channel', 'Connected signal', 'Num. overflows']:
            page.th(header)
        page.thead.close()
        # write body
        page.tbody()
        for c, seglist in overflows.items():
            t = abs(seglist.active)
            if t == 0:
                page.tr()
            elif t == abs(cachesegs):
                page.tr(class_='table-warning')
            else:
                page.tr(class_='table-danger')
            page.td(c)
            try:
                page.td(cds.get_real_channel(str(c)))
            except Exception:
                page.td()
            page.td(len(seglist.active))
            page.tr.close()
        page.tbody.close()
        page.table.close()

        # -- close and write
        htmlio.close_page(page, args.html)
        LOGGER.info("HTML written to %s" % args.html)


# -- run from command-line ----------------------------------------------------

if __name__ == "__main__":
    main()



# coding=utf-8
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Command-line utility for providing status updates to Nagios
"""

import os
import sys

from .. import cli
from .core import write_status

__author__ = 'Alex Urban <alexander.urban@ligo.org>'

PROG = ('python -m gwdetchar.nagios' if sys.argv[0].endswith('.py')
        else os.path.basename(sys.argv[0]))
LOGGER = cli.logger(name=PROG.split('python -m ').pop())


# -- parse command-line -------------------------------------------------------

def create_parser():
    """Create a command-line parser for this entry point
    """
    # initialize argument parser
    parser = cli.create_parser(
        prog=PROG,
        description=__doc__,
    )

    # required arguments
    parser.add_argument(
        'code',
        type=int,
        help='exit code for Nagios, should be one of: '
             '0 (success), 1 (warning), or 2 (critical)',
    )
    parser.add_argument(
        'message',
        type=str,
        help='status message for Nagios',
    )

    # optional arguments
    parser.add_argument(
        '-t',
        '--timeout',
        type=int,
        default=0,
        help='timeout length in seconds, default: no timeout',
    )
    parser.add_argument(
        '-m',
        '--timeout-message',
        default='Process timed out',
        help='Error message upon timeout',
    )
    parser.add_argument(
        '-o',
        '--output-file',
        default='nagios.json',
        help='full path to the output JSON file, '
             'default: %(default)s',
    )

    # return the argument parser
    return parser


# -- main code block ----------------------------------------------------------

def main(args=None):
    """Run the command-line Nagios status generator
    """
    parser = create_parser()
    args = parser.parse_args(args=args)

    # write nagios file
    write_status(
        args.message,
        args.code,
        timeout=args.timeout,
        tmessage=args.timeout_message,
        nagiosfile=args.output_file,
    )

    # log output file path
    LOGGER.info('Status written to {}'.format(
        os.path.abspath(args.output_file)))


# -- run from command-line ----------------------------------------------------

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwdetchar.plot`
"""

import os
import shutil
import warnings

from unittest.mock import patch

from gwpy.segments import DataQualityFlag

from matplotlib import (
    use,
    rcParams,
    MatplotlibDeprecationWarning,
)
use('Agg')

# backend-dependent import
from .. import plot  # noqa: E402

__author__ = 'Alex Urban <alexander.urban@ligo.org>'


# -- test utilities -----------------------------------------------------------

def test_texify():
    name = 'X1:TEST-CHANNEL_NAME'

    # ignore deprecation warnings for rcParams
    warnings.simplefilter('ignore', MatplotlibDeprecationWarning)

    # test with LaTeX
    with patch.dict(rcParams, {'text.usetex': True}):
        assert plot.texify(name) == name.replace('_', r'\_')

    # test without LaTeX
    with patch.dict(rcParams, {'text.usetex': False}):
        assert plot.texify(name) == name

    # null use case
    assert plot.texify(None) == ''


# -- make sure plots run end-to-end -------------------------------------------

def test_plot_segments(tmpdir):
    base = str(tmpdir)
    flag = DataQualityFlag(
        known=[(0, 66)],
        active=[(16, 42)],
        name='X1:TEST-FLAG:1',
    )
    segplot = plot.plot_segments(flag, span=(0, 66))
    segplot.savefig(os.path.join(base, 'test.png'))
    shutil.rmtree(base, ignore_errors=True)



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2020)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the `gwdetchar.omega` command-line interface
"""

import os
import numpy
import pytest
import shutil

from scipy.signal import gausspulse
from unittest import mock

from gwpy.segments import (
    Segment,
    SegmentList,
    DataQualityFlag,
)
from gwpy.timeseries import (
    TimeSeries,
    TimeSeriesDict,
)

from .. import __main__ as omega_cli

__author__ = 'Alex Urban <alexander.urban@ligo.org>'

# -- test configurations

K1_CONFIG = u"""[primary]
f-low = 10.0
source = {path}
duration = 32
fftlength = 8
matched-filter-length = 6
channel = K1:GW-PRIMARY_CHANNEL

[GW]
name = Gravitational Wave Strain
q-range = 3.3166,150.0
frequency-range = 10.0,1024.0
source = {path}
state-flag = K1:DCH-TEST_FLAG:1
duration = 32
fftlength = 8
max-mismatch = 0.2
snr-threshold = 5
always-plot = True
plot-time-durations = 1,4,16
channels = K1:GW-PRIMARY_CHANNEL

[AUX]
name = Auxiliary Sensors
q-range = 3.3166,100.0
frequency-range = 4.0,1024.0
source = {path}
state-flag = K1:DCH-TEST_FLAG:1
duration = 32
fftlength = 8
max-mismatch = 0.35
snr-threshold = 5.5
always-plot = False
plot-time-durations = 1,4,16
channels = K1:AUX-HIGH_SIGNIFICANCE
    K1:AUX-LOW_SIGNIFICANCE
    K1:AUX-INVALID_DATA
""".rstrip()

NETWORK_CONFIG = u"""[H1]
name = LIGO-Hanford
q-range = 3.3166,150.0
frequency-range = 10.0,1024.0
source = {path}
duration = 32
fftlength = 8
max-mismatch = 0.2
snr-threshold = 5
always-plot = True
plot-time-durations = 1,4,16
channels = H1:GW-PRIMARY_CHANNEL

[L1]
name = LIGO-Livingston
q-range = 3.3166,150.0
frequency-range = 10.0,1024.0
source = {path}
duration = 32
fftlength = 8
max-mismatch = 0.2
snr-threshold = 5
always-plot = True
plot-time-durations = 1,4,16
channels = L1:GW-PRIMARY_CHANNEL
""".rstrip()

NETWORK_CONFIG_WITH_PRIMARY = """[primary]
f-low = 10.0
source = {path}
duration = 32
fftlength = 8
matched-filter-length = 6
channel = L1:GW-PRIMARY_CHANNEL

""" + NETWORK_CONFIG

MOCK_SUMMARY = """Channel,Central Time,Central Frequency (Hz),Q,Energy,SNR,Correlation,Standard Deviation,Delay (ms)
H1:GW-PRIMARY_CHANNEL,17.0,69.0,8.6,3021.39990234375,77.7,3.5,1.224587425297728,-50
L1:GW-PRIMARY_CHANNEL,17.0,78.6,8.6,31.200000762939453,7.9,6.3,1.3387611859727206,19
K1:GW-PRIMARY_CHANNEL,17.0,88.0,8.6,70.9000015258789,11.9,221.9,16.73187955362334,0
K1:AUX-HIGH_SIGNIFICANCE,17.0,62.4,8.6,43.400001525878906,9.3,6.9,1.7532027279924842,-50
K1:AUX-LOW_SIGNIFICANCE,17.0,102.7,44.2,33.79999923706055,8.2,13.7,4.709735874989297,2
K1:AUX-INVALID_DATA,17.0,103.0,69.0,103.5999984741211,14.4,5.0,1.445966601520126,-55
"""  # noqa: E501

# -- test data

GPS = 17

SIGNAL = TimeSeries(
    gausspulse(numpy.arange(-1, 1, 1./4096), bw=100),
    sample_rate=4096,
    epoch=GPS - 1,
)

TEST_FLAG = DataQualityFlag(
    name="K1:DCH-TEST_FLAG:1",
    active=SegmentList(),
    known=SegmentList([Segment(0, 34)]),
)

K1_DATA = TimeSeriesDict({
    "K1:GW-PRIMARY_CHANNEL": TimeSeries(
        numpy.random.normal(loc=1, scale=.5, size=4096 * GPS * 2),
        sample_rate=4096,
        epoch=0,
    ).zpk([], [0], 1).inject(SIGNAL),
    "K1:AUX-HIGH_SIGNIFICANCE": TimeSeries(
        numpy.random.normal(loc=1, scale=.5, size=4096 * GPS * 2),
        sample_rate=4096,
        epoch=0,
    ).zpk([], [0], 1).inject(SIGNAL),
    "K1:AUX-LOW_SIGNIFICANCE": TimeSeries(
        numpy.random.normal(loc=1, scale=.5, size=4096 * GPS * 2),
        sample_rate=4096,
        epoch=0,
    ),
    "K1:AUX-INVALID_DATA": TimeSeries(
        numpy.full(4096 * GPS * 2, numpy.nan),
        sample_rate=4096,
        epoch=0,
    ),
})

NETWORK_DATA = TimeSeriesDict({
    "H1:GW-PRIMARY_CHANNEL": TimeSeries(
        numpy.random.normal(loc=1, scale=.5, size=4096 * GPS * 2),
        sample_rate=4096,
        epoch=0,
    ).zpk([], [0], 1).inject(SIGNAL),
    "L1:GW-PRIMARY_CHANNEL": TimeSeries(
        numpy.random.normal(loc=1, scale=.5, size=4096 * GPS * 2),
        sample_rate=4096,
        epoch=0,
    ).zpk([], [0], 1).inject(SIGNAL),
})


# -- utils --------------------------------------------------------------------

def _get_inputs(workdir, config, data):
    """Prepare and return paths to input data products for Omega scan testing
    """
    # get path to data files
    ini_target = os.path.join(workdir, "config.ini")
    data_target = os.path.abspath(os.path.join(workdir, "data.h5"))
    # write to data files and return
    with open(ini_target, 'w') as f:
        f.write(config.format(path=data_target))
    data.write(data_target, format="hdf5", overwrite=True)
    return ini_target


# -- cli tests ----------------------------------------------------------------

def test_main_single_ifo(caplog, tmpdir):
    outdir = str(tmpdir)
    ini_source = _get_inputs(outdir, K1_CONFIG, K1_DATA)
    args = [
        str(GPS),
        '--ifo', 'K1',
        '--config-file', ini_source,
        '--output-dir', outdir,
        '--ignore-state-flags',
    ]
    # test from scratch
    with pytest.warns(UserWarning) as record:
        omega_cli.main(args)
    assert 'K1 Omega Scan {:.1f}'.format(GPS) in caplog.text
    assert caplog.text.count('Channel not significant') == 1
    assert os.path.isfile(os.path.join(outdir, 'index.html'))
    assert os.path.isfile(os.path.join(outdir, 'data', 'summary.csv'))
    assert 'array must not contain infs or NaNs' in record[-1].message.args[0]
    # test with checkpointing
    caplog.clear()
    with pytest.warns(UserWarning) as record:
        omega_cli.main(args)
    assert caplog.text.count('Checkpointing from {}'.format(outdir)) == 1
    assert caplog.text.count('Checkpointing K1:GW-PRIMARY_CHANNEL '
                             'from a previous run') == 1
    assert caplog.text.count('Checkpointing K1:AUX-HIGH_SIGNIFICANCE '
                             'from a previous run') == 1
    assert caplog.text.count('Channel not significant') == 1
    assert 'array must not contain infs or NaNs' in record[-1].message.args[0]
    # clean up
    shutil.rmtree(outdir, ignore_errors=True)


def test_main_multi_ifo(caplog, tmpdir):
    outdir = str(tmpdir)
    ini_source = _get_inputs(outdir, NETWORK_CONFIG, NETWORK_DATA)
    args = [
        str(GPS),
        '--ifo', 'Network',
        '--config-file', ini_source,
        '--output-dir', outdir,
    ]
    # test from scratch
    omega_cli.main(args)
    assert 'Network Omega Scan {}'.format(GPS) in caplog.text
    assert os.path.isfile(os.path.join(outdir, 'index.html'))
    assert os.path.isfile(os.path.join(outdir, 'data', 'summary.csv'))
    # test with checkpointing and --disable-correlation
    ini_source = _get_inputs(
        outdir, NETWORK_CONFIG_WITH_PRIMARY, NETWORK_DATA)
    omega_cli.main(args + ['--disable-correlation'])
    assert 'Checkpointing from {}'.format(outdir) in caplog.text
    # test with checkpointing that expects cross-correlation
    with pytest.raises(KeyError) as exc:
        omega_cli.main(args)
    assert 'Cross-correlation is not available from this record' in str(exc)
    # clean up
    shutil.rmtree(outdir, ignore_errors=True)


@mock.patch('gwpy.segments.DataQualityFlag.query',
            return_value=TEST_FLAG)
def test_main_inactive_segments(segserver, caplog):
    outdir = 'null-test'
    # write input data products to current working directory
    # this is done to test Omega scans' relative path construction
    ini_source = _get_inputs(os.path.curdir, K1_CONFIG, K1_DATA)
    args = [
        str(GPS),
        '--ifo', 'K1',
        '--config-file', ini_source,
        '--output-dir', outdir,
    ]
    omega_cli.main(args)
    # test output
    assert 'Finalizing HTML at {}'.format(
        os.path.abspath(outdir)) in caplog.text
    with open(os.path.join(outdir, "index.html"), 'r') as f:
        assert ('No significant channels found during '
                'active analysis segments') in f.read()
    # clean up
    os.remove('config.ini')
    os.remove('data.h5')
    shutil.rmtree(outdir, ignore_errors=True)


def test_main_invalid_config_files():
    ini_source = '/does/not/exist.ini'
    args = [
        str(GPS),
        '--ifo', 'V1',
        '--config-file', ini_source,
    ]
    # test that a non-existing configuration file raises OSError
    with pytest.raises(OSError) as exc:
        omega_cli.main(args)
    assert str(exc.value) == "Cannot read file '{}'".format(ini_source)


def test_main_multiple_config_files(tmpdir):
    outdir = str(tmpdir)
    os.mkdir(os.path.join(outdir, 'data'))
    ini1 = _get_inputs(outdir, K1_CONFIG, K1_DATA)
    ini2 = os.path.join(outdir, 'network-config.ini')
    with open(ini2, 'w') as f:
        f.write(K1_CONFIG.format(path='data.h5'))
    with open(os.path.join(outdir, 'data', 'summary.csv'), 'w') as f:
        f.write(MOCK_SUMMARY)
    args = [
        str(GPS),
        '--ifo', 'Network',
        '--config-file', ini1,
        '--config-file', ini2,
        '--output-dir', outdir,
        '--ignore-state-flags',
    ]
    # test output
    omega_cli.main(args)
    assert os.path.exists(os.path.join(outdir, 'index.html'))
    # clean up
    shutil.rmtree(outdir, ignore_errors=True)



# coding=utf-8
# Copyright (C) Duncan Macleod (2018)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""HTCondor utilities for `gwdetchar`
"""

import re

from . import const

OBS_RUN_REGEX = re.compile('[OS][0-9]*', re.I)
ACCOUNTING_GROUPS_FILE = '/etc/condor/accounting/valid_tags'


def accounting_epoch(gpstime):
    """Return the appropriate epoch for condor accounting tags

    Examples
    --------
    >>> from gwdetchar.condor import accounting_epoch
    >>> print(accounting_epoch(1126259462))  # GW150914
    'O1'
    """
    for epoch in filter(OBS_RUN_REGEX.match, const.EPOCH):
        # if gpstime falls before the _end_ of this run, use it
        if float(gpstime) < const.EPOCH[epoch][1]:
            return epoch
    # gpstime falls _after_ the end of the last known epoch, use it
    return epoch


def is_valid(tag, path=ACCOUNTING_GROUPS_FILE):
    """Determine whether an accounting tag is valid, and raise an exception
    if not
    """
    try:
        valid = validate_accounting_tag(tag, path=path)
    except EnvironmentError:
        valid = True  # failed to load condor tags, not important
    if not valid:
        listtags = f'cat {path}'
        raise ValueError("condor accounting tag {0!r} not recognised, to see "
                         "the list of valid groups, please run `{1}`".format(
                             tag, listtags))
    return valid


def validate_accounting_tag(tag, path=ACCOUNTING_GROUPS_FILE):
    """Validate a given accounting tag

    This loads the list of known accounting tags from ``path`` and
    checks for ``tag`` in that list.

    Returns
    -------
    True
        if ``tag`` is found in the list of tags
    False
        if ``tag`` is **not** found in the list of tags
    """
    return tag in load_accounting_tags(path=path)


def load_accounting_tags(path=ACCOUNTING_GROUPS_FILE):
    """Load the list of known accounting tags
    """

    # The file is lines like
    # * <tag> <found>\n
    with open(path, 'r') as fobj:
        lines = fobj.readlines()

    for n, line in enumerate(lines):
        lines[n] = line.split(' ')[1]

    return lines



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwdetchar.omega.plot`
"""

import os
import shutil
import tempfile

import numpy
from scipy import signal

from gwpy.timeseries import TimeSeries

from .. import (config, core)

from matplotlib import use
use('agg')  # noqa

# backend-dependent import
from .. import plot  # noqa: E402

__author__ = 'Alex Urban <alexander.urban@ligo.org>'


# global test objects

FFTLENGTH = 8

NOISE = TimeSeries(
    numpy.random.normal(loc=1, scale=.5, size=16384 * 68),
    sample_rate=16384, epoch=-34).zpk([], [0], 1)
GLITCH = TimeSeries(
    signal.gausspulse(numpy.arange(-1, 1, 1./16384), bw=100),
    sample_rate=16384, epoch=-1) * 1e-4
INPUT = NOISE.inject(GLITCH)

CONFIGURATION = {
    'q-range': '3.3166,150',
    'frequency-range': '4.0,Inf',
    'plot-time-durations': '4',
    'always-plot': 'True',
}
CHANNEL = config.OmegaChannel(
    channelname='L1:TEST-STRAIN', section='test', **CONFIGURATION)

SERIES = core.scan(
    gps=0, channel=CHANNEL, xoft=INPUT, resample=2048, fftlength=FFTLENGTH)
XOFT, _, _, QGRAM, _, QSPEC, _ = SERIES


# -- make sure plots run end-to-end -------------------------------------------

def test_timeseries_plot():
    with tempfile.NamedTemporaryFile(suffix='.png') as png:
        plot.timeseries_plot(XOFT, gps=0, span=4, channel=CHANNEL.name,
                             output=png, ylabel='Test')


def test_spectrogram_plot():
    with tempfile.NamedTemporaryFile(suffix='.png') as png:
        plot.spectral_plot(
            QSPEC, gps=0, span=4, channel=CHANNEL.name, output=png)


def test_qgram_plot():
    with tempfile.NamedTemporaryFile(suffix='.png') as png:
        plot.spectral_plot(QGRAM.table(), gps=0, span=4,
                           channel=CHANNEL.name, output=png)


def test_write_qscan_plots(tmpdir):
    tmpdir.mkdir('plots')
    wdir = str(tmpdir)
    os.chdir(wdir)
    plot.write_qscan_plots(gps=0, channel=CHANNEL, series=SERIES)
    shutil.rmtree(wdir, ignore_errors=True)



# coding=utf-8
# Copyright (C) Duncan Macleod (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Pytest configuration for GWDetChar
"""

import os
from unittest import mock

import pytest


def fake_package_list():
    """Fake return for `gwdetchar.io.html.package_list`

    The real :func:`package_list` takes a while and doesn't add
    anything to any of the tests, so we should fake it almost all
    of the time.
    """
    return [
        {
            "name": "package1",
            "version": "1.0.0",
            "build_string": "0",
            "channel": "conda-forge",
        },
        {
            "name": "package2",
            "version": "2.0.0",
            "build_string": "1",
            "channel": "conda-forge",
        },
    ]


@pytest.fixture(autouse=True)
def mock_package_list(request):
    """Automatically mock out `gwdetchar.io.html.package_list` in all tests
    """
    # don't apply the mock to functions that test `package_list` itself
    if "package_list" in request.node.name:
        yield
        return
    # mock everything else
    with mock.patch("gwdetchar.io.html.package_list", fake_package_list):
        yield


# -- pytest fixture overrides -------------------------------------------------
# these new fixtures overload the pytest builtin `tmpdir` and `tmp_path`
# to ensure that the session is returned to the starting directory once
# the test has finished (regardless of state).  This is a quick-fix to
# address the fact that many existing tests execute `os.chdir(tmpdir)`
# or similar, then run `shutil.rmtree`, leaving the session with a CWD
# that doesn't exist.

@pytest.fixture
def tmpdir(tmpdir):
    """Overload pytest's `tmpdir` to preserve the CWD from the test start
    """
    start = os.getcwd()
    try:
        yield tmpdir
    finally:
        os.chdir(start)


@pytest.fixture
def tmp_path(tmp_path):
    """Overload pytest's `tmp_path` to preserve the CWD from the test start
    """
    start = os.getcwd()
    try:
        yield tmp_path
    finally:
        os.chdir(start)



# coding=utf-8
# Copyright (C) Duncan Macleod (2015)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Utilties for LIGO_LW XML I/O
"""

try:
    from glue.ligolw import (ligolw, table, lsctables)
except ImportError as exc:
    exc.args = ("{!s}, please install lscsoft-glue to "
                "handle LIGO_LW files".format(exc),)
    raise

from gwpy.segments import (Segment, DataQualityFlag, DataQualityDict)

LIGOTimeGPS = lsctables.LIGOTimeGPS

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def new_table(tab, *args, **kwargs):
    """Create a new `~glue.ligolw.Table`

    This is just a convenience wrapper around `~glue.ligolw.lsctables.New`

    Parameters
    ----------
    tab : `type`, `str`
        `~glue.ligolw.Table` subclass, or name of table to create
    *args, **kwargs
        other parameters are passed directly to `~glue.ligolw.lsctables.New`

    Returns
    -------
    table : `~glue.ligolw.Table`
        a newly-created table with the relevant attributes and structure
    """
    if isinstance(tab, str):
        tab = lsctables.TableByName[table.Table.TableName(tab)]
    return lsctables.New(tab, *args, **kwargs)


def sngl_burst_from_times(times, **params):
    """Create a `SnglBurstTable` from an array of times
    """
    columns = set(params.keys()) | {'peak_time', 'peak_time_ns', 'event_id'}
    table = new_table('sngl_burst', columns=list(columns))
    get_next_id = table.get_next_id
    RowType = table.RowType
    append = table.append
    for t in times:
        row = RowType()
        row.event_id = get_next_id()
        row.set_peak(LIGOTimeGPS(t))
        for key, val in params.items():
            setattr(row, key, val)
        append(row)
    return table


def sngl_burst_from_segments(segs, **params):
    """Create a `SnglBurstTable from a `~ligo.segments.segmentlist`
    """
    table = new_table('sngl_burst', columns=params.keys())
    next_id = table.next_id
    RowType = table.RowType
    append = table.append
    for seg in segs:
        row = RowType()
        row.event_id = next_id()
        row.set_period(seg)
        row.set_peak(LIGOTimeGPS(seg[0] + abs(seg)/2.))
        for key, val in params.items():
            setattr(row, key, val)
        append(row)
    return table


def segments_from_sngl_burst(table, padding, known=None):
    """Create a `DataQualityDict` of segments from the given `SnglBurstTable`

    This method creates a `~gwpy.segments.DataQualityFlag` for each unique
    channel found in the given table by padding the peak time by the given
    amount on each side
    """
    out = DataQualityDict()
    for row in table:
        t = row.get_peak()
        seg = Segment(t-padding, t+padding)
        try:
            out[row.channel].active.append(seg)
        except KeyError:
            out[row.channel] = DataQualityFlag(row.channel, known=known,
                                               active=[seg])
    return out


def table_to_document(table):
    xmldoc = ligolw.Document()
    xmldoc.appendChild(ligolw.LIGO_LW())
    xmldoc.childNodes[0].appendChild(table)
    return xmldoc



# coding=utf-8
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for providing status updates to Nagios
"""

import json

from gwpy.time import to_gps

__author__ = 'Alex Urban <alexander.urban@ligo.org>'


# -- utilities ----------------------------------------------------------------

def write_status(message, code, timeout=0, tmessage=None, nagiosfile=None):
    """Write a Nagios status file in JSON format

    Parameters
    ----------
    message : `str`
        status message for Nagios

    code : `int`
        exit code for process

    timeout : `int`, optional
        timeout length, in seconds

    tmessage : `str`, optional
        timeout message

    nagiosfile : `str`, optional
        full path to a JSON status file, defaults to ``nagios.json``

    Notes
    -----
    This function will write an output to the requested location, then exit
    without returning.
    """
    # status dictionary
    status = {
        "created_gps": int(to_gps('now')),
        "status_intervals": [{
            "start_sec": 0,
            "txt_status": message,
            "num_status": code
        }],
    }
    # update timeout information
    if timeout:
        status["status_intervals"].append({
            "start_sec": timeout,
            "txt_status": tmessage,
            "num_status": 3,
        })
    # get output file and write
    nagiosfile = nagiosfile or 'nagios.json'
    with open(nagiosfile, 'w') as fileobj:
        json.dump(status, fileobj)



# coding=utf-8
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Methods and utilties for optical scattering
"""

__author__ = 'Alex Urban <alexander.urban@ligo.org>'
__credits__ = 'Duncan Macleod <duncan.macleod@ligo.org>' \
              'Joshua Smith <joshua.smith@ligo.org>' \
              'Andrew Lundgren <andrew.lundgren>@ligo.org>'

# import scattering utils
from .core import (
    OPTIC_MOTION_CHANNELS,
    TRANSMON_CHANNELS,
    FREQUENCY_MULTIPLIERS,
    get_fringe_frequency,
    get_blrms,
    get_segments,
)

# global figure captions
SCATTER_CAPTION = 'Evidence for scattering in {CHANNEL}. Top: longitudinal ' \
                  'optic motion; second row: projected fringe frequency of ' \
                  'the first four harmonics; third row: time-frequency ' \
                  'scatter plot of Omicron triggers, coloured by signal-to-' \
                  'noise ratio; bottom: known (small) and active (large) ' \
                  'time segments when one harmonic is above the frequency ' \
                  'threshold.'
HIST_CAPTION = 'Cumulative histogram of the total time with fringe ' \
               'frequency above a given value. The first (red), second ' \
               '(green), third (gold), and fourth (blue) harmonics are ' \
               'shown for {CHANNEL}.'



# coding=utf-8
# Copyright (C) Alex Urban (2018-2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Plotting routines for scattering checks
"""

from gwpy.plot import Plot

__author__ = 'Alex Urban <alexander.urban@ligo.org>'
__credits__ = 'Joshua Smith <joshua.smith@ligo.org>' \
              'Andrew Lundgren <andrew.lundgren>@ligo.org>'


# -- custom plotting tools ----------------------------------------------------

def _format_timeseries(ax, gps, fringe, multipliers=(1, 2, 4, 8),
                       linewidth=1, thresh=None):
    """Helper tool to format a `~gwpy.timeseries.TimeSeries` plot axis
    """
    t1 = fringe.t0.value
    t2 = (fringe.t0 + fringe.duration).value
    if thresh is not None:
        ax.plot([t1, t2], [thresh, thresh], 'k--')
    for m in sorted(multipliers, reverse=True):
        ax.plot(m * fringe, label=r'$f\times %d$' % m, linewidth=linewidth)
    # format time axis
    ax.set_xlim([t1, t2])
    ax.set_xscale('seconds', epoch=gps)
    ax.grid(True, axis='x', which='major')
    ax.legend(loc='upper right')


def _format_spectrogram(ax, qspecgram, colormap='viridis'):
    """Helper tool to format a `~gwpy.spectrogram.Spectrogram` plot axis
    """
    ax.imshow(qspecgram)
    ax.colorbar(cmap=colormap, norm='linear', clim=(0, 25),
                label='Normalized energy')
    # format frequency axis
    ax.set_yscale('linear')
    ax.set_ylabel('Frequency [Hz]')
    ax.grid(True, axis='y', which='both')


def spectral_comparison(gps, qspecgram, fringe, output, thresh=15,
                        multipliers=(1, 2, 4, 8), colormap='viridis',
                        figsize=[12, 8]):
    """Compare a high-resolution spectrogram with projected fringe frequencies

    Parameters
    ----------
    gps : `float`
        reference GPS time (in seconds) to serve as the origin

    qspecgram : `~gwpy.spectrogram.Spectrogram`
        an interpolated high-resolution spectrogram

    fringe : `~gwpy.timeseries.TimeSeries`
        projected fringe frequencies (in Hz)

    output : `str`
        name of the output file

    thresh : `float`, optional
        frequency threshold (Hz) for scattering fringes, default: 15

    multipliers : `tuple`, optional
        collection of fringe harmonic numbers to plot, can be given in
        any order, default: `(1, 2, 4, 8)`

    colormap : `str`, optional
        matplotlib colormap to use, default: viridis

    figsize : `tuple`, optional
        size (width x height) of the final figure, default: `(12, 8)`
    """
    plot = Plot(figsize=figsize)
    # format spectral plot
    ax1 = plot.add_subplot(211)
    ax1.set_title('{0} at {1:.2f} with $Q$ of {2:.1f}'.format(
        qspecgram.name, gps, qspecgram.q))
    _format_spectrogram(ax1, qspecgram, colormap=colormap)
    ax1.set_xlabel(None)
    # format fringe frequency plot
    ax2 = plot.add_subplot(212, sharex=ax1)
    ax2.set_title('{} scattering fringes'.format(fringe.name))
    _format_timeseries(ax2, gps, fringe, multipliers=multipliers,
                       thresh=thresh)
    # format timeseries axes
    ax2.set_ylim([-1, 60])
    ax2.set_ylabel('Projected Frequency [Hz]')
    # save plot and close
    plot.savefig(output, bbox_inches='tight')
    plot.close()


def spectral_overlay(gps, qspecgram, fringe, output,
                     multipliers=(1, 2, 4, 8), figsize=[12, 4]):
    """Overlay scattering fringe projections on top of a high-resolution
    spectrogram

    Parameters
    ----------
    gps : `float`
        reference GPS time (in seconds) to serve as the origin

    qspecgram : `~gwpy.spectrogram.Spectrogram`
        an interpolated high-resolution spectrogram

    fringe : `~gwpy.timeseries.TimeSeries`
        projected fringe frequencies (in Hz)

    output : `str`
        name of the output file

    multipliers : `tuple`, optional
        collection of fringe harmonic numbers to plot, can be given in
        any order, default: `(1, 2, 4, 8)`

    figsize : `tuple`, optional
        size (width x height) of the final figure, default: `(12, 4)`
    """
    plot = Plot(figsize=figsize)
    ax = plot.gca()
    # format spectrogram plot
    ax.set_title('Fringes: {0}, Spectrogram: {1}'.format(
        fringe.name, qspecgram.name))
    _format_spectrogram(ax, qspecgram, colormap='binary')
    # overlay fringe frequencies
    _format_timeseries(ax, gps, fringe, multipliers=multipliers,
                       linewidth=1.5)
    ax.set_ylim([qspecgram.f0.to('Hz').value,
                 qspecgram.frequencies.max().to('Hz').value])
    # save plot and close
    plot.savefig(output, bbox_inches='tight')
    plot.close()



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwdetchar.omega.config`
"""

import os
from io import StringIO

import numpy
from scipy import signal

from gwpy.table import Table
from gwpy.timeseries import TimeSeries

from .. import (config, core)
from ...io.html import FancyPlot

__author__ = 'Alex Urban <alexander.urban@ligo.org>'


# global test objects

CONFIGURATION = u"""
[primary]
; the primary channel, which will be used as a matched-filter
f-low = 4.0
resample = 4096
frametype = X1_HOFT
duration = 64
fftlength = 8
matched-filter-length = 6
channel = X1:TEST-STRAIN

[GW]
; name of this block, which contains h(t)
name = Gravitational Wave Strain
q-range = 3.3166,150.0
frequency-range = 4.0,1024
resample = 4096
frametype = X1_HOFT
state-flag = X1:OBSERVING:1
duration = 64
fftlength = 8
max-mismatch = 0.2
snr-threshold = 5
always-plot = True
plot-time-durations = 4
channels = X1:TEST-STRAIN
"""
CONFIG_FILE = StringIO(CONFIGURATION)

CP = config.OmegaConfigParser(ifo='X1')
CP.read_file(CONFIG_FILE)
BLOCKS = CP.get_channel_blocks()
PRIMARY = BLOCKS['primary']
GW = BLOCKS['GW']

ROWS = [[0]] * 8
COLS = ('Q', 'Energy', 'SNR', 'Central Time', 'Central Frequency (Hz)',
        'Correlation', 'Standard Deviation', 'Delay (ms)')
TABLE = Table(ROWS, names=COLS)


# -- test utilities -----------------------------------------------------------

def test_get_default_configuration():
    cfile = config.get_default_configuration(ifo='X1', gpstime=1126259462)
    assert cfile == [os.path.expanduser(
        '~detchar/etc/omega/ER8/X-X1_R-selected.ini')]
    nfile = config.get_default_configuration(ifo='Network', gpstime=1187008882)
    assert nfile == [os.path.expanduser('~detchar/etc/omega/O2/Network.ini')]


def test_get_fancyplots():
    fp = config.get_fancyplots(
        channel='X1:TEST-STRAIN', plottype='test-plot', duration=4)
    fname = 'plots/X1-TEST_STRAIN-test-plot-4.png'
    assert isinstance(fp, FancyPlot)
    assert fp.img == fname
    assert str(fp) == fp.img
    assert fp.caption == os.path.basename(fname)


def test_config_parser():
    # basic tests
    assert isinstance(CP, config.OmegaConfigParser)
    assert CP.sections() == ['primary', 'GW']

    # test primary section params
    assert CP.get('primary', 'f-low') == '4.0'
    assert CP.get('primary', 'resample') == '4096'
    assert CP.get('primary', 'frametype') == 'X1_HOFT'
    assert CP.get('primary', 'duration') == '64'
    assert CP.get('primary', 'fftlength') == '8'
    assert CP.get('primary', 'matched-filter-length') == '6'
    assert CP.get('primary', 'channel') == 'X1:TEST-STRAIN'

    # test GW section params
    assert CP.get('GW', 'name') == 'Gravitational Wave Strain'
    assert CP.get('GW', 'q-range') == '3.3166,150.0'
    assert CP.get('GW', 'frequency-range') == '4.0,1024'
    assert CP.get('GW', 'resample') == '4096'
    assert CP.get('GW', 'frametype') == 'X1_HOFT'
    assert CP.get('GW', 'state-flag') == 'X1:OBSERVING:1'
    assert CP.get('GW', 'duration') == '64'
    assert CP.get('GW', 'fftlength') == '8'
    assert CP.get('GW', 'max-mismatch') == '0.2'
    assert CP.get('GW', 'snr-threshold') == '5'
    assert CP.get('GW', 'always-plot') == 'True'
    assert CP.get('GW', 'plot-time-durations') == '4'
    assert CP.get('GW', 'channels') == 'X1:TEST-STRAIN'


def test_omega_channel_list():
    # basic test
    assert isinstance(GW, config.OmegaChannelList)

    # test primary block
    assert PRIMARY.key == 'primary'
    assert PRIMARY.parent is None
    assert PRIMARY.name is None
    assert PRIMARY.duration == 64
    assert PRIMARY.fftlength == 8
    assert PRIMARY.resample == 4096
    assert PRIMARY.source is None
    assert PRIMARY.frametype == 'X1_HOFT'
    assert PRIMARY.length == 6
    assert PRIMARY.flow == 4.
    assert PRIMARY.channel == config.OmegaChannel(
        'X1:TEST-STRAIN', PRIMARY.key, **PRIMARY.params)

    # test GW block
    assert GW.key == 'GW'
    assert GW.parent is None
    assert GW.name == 'Gravitational Wave Strain'
    assert GW.duration == 64
    assert GW.fftlength == 8
    assert GW.resample == 4096
    assert GW.source is None
    assert GW.frametype == 'X1_HOFT'
    assert GW.flag == 'X1:OBSERVING:1'
    assert GW.search == 0.5
    assert GW.dt == 0.1
    assert GW.channels == [config.OmegaChannel(
        'X1:TEST-STRAIN', GW.key, **GW.params)]


def test_omega_channel():
    # basic test
    channel = GW.channels[0]
    assert isinstance(channel, config.OmegaChannel)

    # test attributes
    assert channel.section == 'GW'
    assert channel.name == 'X1:TEST-STRAIN'
    assert channel.frametype == 'X1_HOFT'
    assert channel.qrange == (3.3166, 150.0)
    assert channel.frange == (4.0, 1024)
    assert channel.mismatch == 0.2
    assert channel.snrthresh == 5
    assert channel.always_plot is True
    assert channel.pranges == [4]


def test_save_loudest_tile_features():
    # prepare input data
    channel = GW.channels[0]
    noise = TimeSeries(
        numpy.random.normal(loc=1, scale=.5, size=16384 * 68),
        sample_rate=16384, epoch=-34).zpk([], [0], 1)
    glitch = TimeSeries(
        signal.gausspulse(numpy.arange(-1, 1, 1./16384), bw=100),
        sample_rate=16384, epoch=-1) * 1e-4
    in_ = noise.inject(glitch)
    _, _, _, qgram, _, _, _ = core.scan(
        gps=0, channel=channel, xoft=in_, resample=4096, fftlength=8)

    # test loudest tiles
    channel.save_loudest_tile_features(qgram, correlate=glitch)
    assert channel.Q == numpy.around(qgram.plane.q, 1)
    assert channel.energy == numpy.around(qgram.peak['energy'], 1)
    assert channel.snr == numpy.around(qgram.peak['snr'], 1)
    assert channel.t == numpy.around(qgram.peak['time'], 3)
    assert channel.f == numpy.around(qgram.peak['frequency'], 1)
    assert channel.corr == numpy.around(glitch.max().value, 1)
    assert channel.delay == 0.0
    assert channel.stdev == glitch.std().value


def test_load_loudest_tile_features():
    channel = GW.channels[0]
    channel.load_loudest_tile_features(TABLE, correlated=True)
    assert channel.Q == 0
    assert channel.energy == 0
    assert channel.snr == 0
    assert channel.t == 0
    assert channel.f == 0
    assert channel.corr == 0
    assert channel.delay == 0
    assert channel.stdev == 0
